# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data models for Tailbone
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa

from rattail.db.model import Base, uuid_column


@six.python_2_unicode_compatible
class TailbonePageHelp(Base):
    """
    Represents help info for a particular page (or more often, set of
    pages) within Tailbone.
    """
    __tablename__ = 'tailbone_page_help'
    __table_args__ = (
        sa.UniqueConstraint('route_prefix',
                            name='tailbone_page_help_uq_route_prefix'),
    )
    __versioned__ = {}

    uuid = uuid_column()

    route_prefix = sa.Column(sa.String(length=254), nullable=False, doc="""
    Route prefix in Tailbone, to which this help info applies.
    """)

    help_url = sa.Column(sa.String(length=254), nullable=True, doc="""
    URL to (probably external) help document.
    """)

    markdown_text = sa.Column(sa.Text(), nullable=True, doc="""
    Help text as markdown.
    """)

    def __str__(self):
        return self.route_prefix or ''


@six.python_2_unicode_compatible
class TailboneFieldInfo(Base):
    """
    Represents info for a particular form field within Tailbone.
    """
    __tablename__ = 'tailbone_field_info'
    __table_args__ = (
        sa.UniqueConstraint('route_prefix', 'field_name',
                            name='tailbone_field_info_uq_field'),
    )
    __versioned__ = {}

    uuid = uuid_column()

    route_prefix = sa.Column(sa.String(length=254), nullable=False, doc="""
    Route prefix in Tailbone, to which this help info applies.
    """)

    field_name = sa.Column(sa.String(length=100), nullable=False, doc="""
    Name of the field within Tailbone code.
    """)

    markdown_text = sa.Column(sa.Text(), nullable=True, doc="""
    Help text as markdown.
    """)

    def __str__(self):
        return self.route_prefix or ''



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Commands related to purging of old data
"""

import os
import datetime
import shutil
import logging

from rattail.commands import Subcommand, date_argument


log = logging.getLogger(__name__)


class PurgeSubcommand(Subcommand):
    """
    Base class for subcommands which purge old data
    """
    purge_title = "Object"
    default_before_days = 90

    @property
    def purge_title_plural(self):
        return "{}s".format(self.purge_title)

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('--before', type=date_argument, metavar='DATE',
                            help="Use this date as cutoff, i.e. purge all data "
                            "*before* this date.  If not specified, will use "
                            "--before-days to calculate instead.")

        parser.add_argument('--before-days', type=int, metavar='DAYS',
                            default=self.default_before_days,
                            help="Calculate the cutoff date by subtracting this "
                            "number of days from the current date, i.e. purge all "
                            "data *before* the resulting date.  Note that if you "
                            "specify --before then that date will be used instead "
                            "of calculating one from --before-days.  If neither is "
                            "specified then --before-days is used, with its default "
                            "value of 90 days.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging "
                            "etc. to occur, but rollback (abort) the transaction "
                            "at the end.")

    def run(self, args):
        run_purge(self.config, self.purge_title, self.purge_title_plural,
                  self.find_things_to_purge, self.purge_thing,
                  before=args.before, before_days=args.before_days,
                  default_before_days=self.default_before_days,
                  dry_run=args.dry_run)

    def find_things_to_purge(self, session, cutoff, dry_run=False):
        pass

    def purge_thing(self, session, thing, cutoff, dry_run=False):
        """
        This method should contain logic which actually "purges" something.
        """
        log.info("purging object: %s", thing)
        return True


class PurgeExport(PurgeSubcommand):
    """
    Base classes for data which purge "export" objects
    """

    @property
    def purge_model_class(self):
        raise NotImplementedError("You must define %s.purge_model_class",
                                  self.__class__.__name__)

    @property
    def purge_title(self):
        return self.purge_model_class.get_model_title()

    @property
    def purge_title_plural(self):
        return self.purge_model_class.get_model_title_plural()

    def find_things_to_purge(self, session, cutoff, dry_run=False):
        model = self.model
        exports = session.query(self.purge_model_class)\
                         .filter(self.purge_model_class.created < self.app.make_utc(cutoff))\
                         .all()
        return exports

    def purge_thing(self, session, export, cutoff, dry_run=False):
        """
        This method does the basics, i.e. deleting the export record from the
        database session, and (if not dry-run) deleting any associated files
        from disk.
        """
        uuid = export.uuid
        log.debug("purging export object %s: %s", uuid, export)
        session.delete(export)

        # maybe delete associated files
        if not dry_run:
            session.flush()
            key = self.purge_model_class.export_key
            path = self.config.export_filepath(key, uuid)
            if os.path.exists(path):
                shutil.rmtree(path)

        return True


def run_purge(config, purge_title, purge_title_plural, thing_finder, thing_purger,
              before=None, before_days=None, default_before_days=90,
              dry_run=False):
    from rattail.db.util import finalize_session

    log.info("will purge things of type: %s", purge_title)

    if before and before_days:
        log.warning("specifying both --before and --before-days is "
                    "redundant; --before will take precedence.")

    app = config.get_app()
    session = app.make_session()

    # calculate our cutoff date
    if before:
        cutoff = before
    else:
        today = app.today()
        cutoff = today - datetime.timedelta(days=before_days or default_before_days)
    cutoff = datetime.datetime.combine(cutoff, datetime.time(0))
    cutoff = app.localtime(cutoff)
    log.info("using %s as cutoff date", cutoff.date())

    # find things, and purge them
    things = thing_finder(session, cutoff, dry_run=dry_run)
    log.info("found %s thing(s) to purge", len(things or []))
    if things:
        purged = purge_things(config, session, things, thing_purger, cutoff, dry_run=dry_run)
        log.info("%spurged %s %s",
                 "(would have) " if dry_run else "",
                 purged, purge_title_plural)

    finalize_session(session, dry_run=dry_run)


def purge_things(config, session, things, purger, cutoff, dry_run=False):
    app = config.get_app()
    result = app.make_object(purged=0)

    def purge(thing, i):
        if purger(session, thing, cutoff, dry_run=dry_run):
            result.purged += 1
        if i % 200 == 0:
            session.flush()

    app.progress_loop(purge, things, progress,
                      message=f"Purging {purge_title_plural}")
    return result.purged



# -*- coding: utf-8 -*-
"""add credit.total

Revision ID: a295dee8d1e9
Revises: a03ee3718ff4
Create Date: 2018-06-27 15:11:54.973213

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'a295dee8d1e9'
down_revision = u'a03ee3718ff4'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch_credit
    op.add_column('purchase_batch_credit', sa.Column('credit_total', sa.Numeric(precision=7, scale=2), nullable=True))

    # purchase_credit
    op.add_column('purchase_credit', sa.Column('credit_total', sa.Numeric(precision=7, scale=2), nullable=True))


def downgrade():

    # purchase_credit
    op.drop_column('purchase_credit', 'credit_total')

    # purchase_batch_credit
    op.drop_column('purchase_batch_credit', 'credit_total')



# -*- coding: utf-8; -*-
"""add department.tax

Revision ID: 7f025fc530cc
Revises: d69dd44f495c
Create Date: 2023-10-11 18:13:57.297032

"""

# revision identifiers, used by Alembic.
revision = '7f025fc530cc'
down_revision = 'd69dd44f495c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # department
    op.add_column('department', sa.Column('tax_uuid', sa.String(length=32), nullable=True))
    op.add_column('department', sa.Column('food_stampable', sa.Boolean(), nullable=True))
    op.create_foreign_key('department_fk_tax', 'department', 'tax', ['tax_uuid'], ['uuid'])
    op.add_column('department_version', sa.Column('tax_uuid', sa.String(length=32), autoincrement=False, nullable=True))
    op.add_column('department_version', sa.Column('food_stampable', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # department
    op.drop_column('department_version', 'food_stampable')
    op.drop_column('department_version', 'tax_uuid')
    op.drop_constraint('department_fk_tax', 'department', type_='foreignkey')
    op.drop_column('department', 'food_stampable')
    op.drop_column('department', 'tax_uuid')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Luigi task commands
"""

import datetime
import sys

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand, date_argument


@rattail_typer.command()
def overnight(
        ctx: typer.Context,
        task_key: Annotated[
            str,
            typer.Option('--task-key', '-k',
                         help="Config key for the overnight task to be launched.")] = None,
        date: Annotated[
            datetime.datetime,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Date for which overnight task should be "
                         "launched.  Defaults to yesterday.")] = None,
        wait: Annotated[
            bool,
            typer.Option(help="Whether to run the task in-process and wait until "
                         "it completes (--wait), or schedule an async run via "
                         "`at` (--no-wait).")] = True,
        email_key: Annotated[
            str,
            typer.Option(help="Config key for email settings, to be used in "
                         "determining recipients etc.")] = None,
        email_if_empty: Annotated[
            bool,
            typer.Option(help="Whether to send email when task produces no output.")] = True,
        purge_settings: Annotated[
            bool,
            typer.Option('--purge-settings',
                         help="Instead of running a task, remove all settings "
                         "from the DB, for all overnight tasks.")] = False,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Log the final command for the task, but do not "
                         "actually run it.")] = False,
):
    """
    Launch an overnight task for Luigi
    """
    config = ctx.parent.rattail_config

    if purge_settings:
        do_purge_settings(config, dry_run=dry_run)
    else:
        do_launch_task(config, ctx.params)


class Overnight(Subcommand):
    """
    Launch an overnight task for Luigi
    """
    name = 'overnight'
    description = __doc__.strip()

    def add_parser_args(self, parser):

        parser.add_argument('--task_key', '-k',
                            help="Config key for the overnight task to be launched.")

        parser.add_argument('--date', type=date_argument,
                            help="Date for which overnight task should be "
                            "launched.  Defaults to yesterday.")

        parser.add_argument('--email-if-empty', action='store_true', default=True,
                            help="Send email even if task produces no output.")
        parser.add_argument('--no-email-if-empty', action='store_false', dest='email_if_empty',
                            help="Send email only if task produces output.")

        parser.add_argument('--email-key',
                            help="Config key for email settings, to be used in "
                            "determining recipients etc.")

        parser.add_argument('--wait', action='store_true', default=True,
                            help="Run the task in-process, and wait for it to "
                            "finish.  This is the default behavior.")
        parser.add_argument('--no-wait', action='store_false', dest='wait',
                            help="Do not wait for the task to finish.  This will "
                            "cause the command to run via the `at` utility, i.e. "
                            "scheduled to start within the next minute.")

        parser.add_argument('--purge-settings', action='store_true',
                            help="Instead of running a task, remove all settings "
                            "from the DB, for all overnight tasks.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Log the final command for the task, but do not "
                            "actually run it.")

    def run(self, args):
        if args.purge_settings:
            do_purge_settings(self.config, dry_run=args.dry_run)
        else:
            params = dict(args._get_kwargs())
            do_launch_task(self.config, params)


def do_launch_task(config, params):
    app = config.get_app()
    luigi = app.get_luigi_handler()
    key = params['task_key']
    task = luigi.get_overnight_task(key)
    if not task:
        sys.stderr.write("overnight task not found for key: {}\n".format(key))
        sys.exit(1)

    date = params['date'] or app.yesterday()
    if isinstance(date, datetime.datetime):
        date = date.date()
    luigi.launch_overnight_task(task, date,
                                email_if_empty=params['email_if_empty'],
                                email_key=params['email_key'],
                                wait=params['wait'],
                                dry_run=params['dry_run'])


def do_purge_settings(config, dry_run=False):
    from rattail.db.util import finalize_session

    app = config.get_app()
    luigi = app.get_luigi_handler()
    session = app.make_session()
    luigi.purge_overnight_settings(session)
    finalize_session(session, dry_run=dry_run)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Project Generators
"""

import os
import random
import re
import shutil
import string
import sys

import colander
from mako.template import Template

from rattail.util import get_class_hierarchy, get_studly_prefix
from rattail.mako import ResourceTemplateLookup


class ProjectGenerator(object):
    """
    Base class for project generators.

    .. attribute:: key

       Unique key for the project type.
    """

    def __init__(self, config, **kwargs):
        self.config = config
        self.app = self.config.get_app()

        self.template_lookup = ResourceTemplateLookup()

    @property
    def key(self):
        raise NotImplementedError("Must define {}.key".format(
            self.__class__.__name__))

    @classmethod
    def get_templates_path(cls):
        """
        Return the path to templates folder for this generator.

        This is a class method for sake of inheritance, so templates
        can more easily be shared by generator subclasses.
        """
        basedir = os.path.dirname(sys.modules[cls.__module__].__file__)
        return os.path.join(basedir, cls.key)

    def make_schema(self, **kwargs):
        return colander.Schema()

    def generate_project(self, output, context, **kwargs):
        """
        Generate a new project to the given output folder, with the
        given context data.

        :param output: Path to output folder.  This path should
           already exist.

        :param context: Dictionary of template context data.
        """

    def get_template_path(self, template):
        """
        Return the full path to a template file.

        :param template: Filename of the template, e.g. ``'setup.py'``.
        """
        for cls in get_class_hierarchy(self.__class__, topfirst=False):
            templates = cls.get_templates_path()
            path = os.path.join(templates, template)
            if os.path.exists(path):
                return path

        raise RuntimeError("template not found: {}".format(template))

    def normalize_context(self, context):
        return context

    def generate(self, template, output, context=None, **kwargs):
        """
        Generate a file from the given template, and save the result
        to the given output path.

        This will do one of 3 things based on the specified template:

        * if filename ends with ``.mako`` then call
          :meth:`generate_mako()`
        * if filename ends with ``.mako_tmpl`` then call
          :meth:`generate_mako_tmpl()`
        * otherwise copy file as-is (do not "generate" output)

        :param template: Path to template file.

        :param output: Path to output file.

        :param context: Data dictionary with template context.
        """
        template = self.get_template_path(template)
        context = context or {}

        # maybe run it through our simplistic, hand-rolled template
        # engine (note, this is only for the sake of *avoiding* mako
        # logic, when generating "actual" mako templates, so we avoid
        # a mako-within-mako situation.)
        if template.endswith('.mako_tmpl'):
            return self.generate_mako_tmpl(template, output, context)

        # maybe run it through Mako template engine
        if template.endswith('.mako'):
            return self.generate_mako(template, output, context)

        # or, just copy the file as-is
        shutil.copyfile(template, output)

    def generate_mako(self, template, output, context):
        """
        Generate output from a Mako template.
        """
        template = Template(filename=template,
                            lookup=self.template_lookup)
        text = template.render(**context)
        with open(output, 'wt') as f:
            f.write(text)

    def generate_mako_tmpl(self, template, output, context):
        """
        Generate output (which is itself a Mako template) from a
        "simple" original template.

        Sometimes you want the final output to be a Mako template, but
        your original template also must be dynamic, based on context.
        It's problematic (confusing at the very least) for an original
        Mako template to produce output which is also a Mako template.
        So instead..

        If you give your original template file a ``.mako_tmpl``
        extension, then the output will still be dynamic, but instead
        of running the original template through the Mako engine, we
        leverage Python strings' printf-style formatting.

        A small example template might be ``rattail.conf.mako_tmpl``:

        .. code-block:: ini

           <%%text>##############################</%%text>
           # example config
           <%%text>##############################</%%text>

           [rattail]
           app_title = ${app_title}

           [alembic]
           script_location = rattail.db:alembic
           version_locations = %(alembic_version_locations)s

           # -- LOGGING SECTION --

           [formatter_generic]
           format = %%(asctime)s %%(levelname)-5.5s [%%(name)s][%%(threadName)s] %%(funcName)s: %%(message)s
           datefmt = %%Y-%%m-%%d %%H:%%M:%%S

        Note the Mako syntax which is *ignored* (passed through as-is)
        by this method when generating output.

        Note also this template expects ``alembic_version_locations``
        to be provided via the context.

        Finally also note the doubled-up ``%`` chars, both in
        ``<%%text>`` as well as in the logging section.  Since
        printf-style formatting uses the ``%`` char, we must escape
        those by doubling-up; the formatter will convert each back to
        single chars for the output.

        For more info on the formatting specifics see
        :ref:`python:old-string-formatting`.

        So with context like::

           {'alembic_version_locations': 'rattail.db:alembic/versions'}

        The above example would produce output ``rattail.conf.mako``:

        .. code-block:: ini

           <%text>##############################</%text>
           # example config
           <%text>##############################</%text>

           [rattail]
           app_title = ${app_title}

           [alembic]
           script_location = rattail.db:alembic
           version_locations = rattail.db:alembic/versions

           # -- LOGGING SECTION --

           [formatter_generic]
           format = %(asctime)s %(levelname)-5.5s [%(name)s][%(threadName)s] %(funcName)s: %(message)s
           datefmt = %Y-%m-%d %H:%M:%S
        """
        with open(template, 'rt') as f:
            template_lines = f.readlines()

        output_lines = []
        for line in template_lines:
            line = line.rstrip('\n')
            line = line % context
            output_lines.append(line)

        with open(output, 'wt') as f:
            f.write('\n'.join(output_lines))

    def random_string(self, size=20, chars=string.ascii_letters + string.digits):
        # per https://stackoverflow.com/a/2257449
        return ''.join(random.SystemRandom().choice(chars) for _ in range(size))


class PythonProjectGenerator(ProjectGenerator):
    """
    Base class for Python project generators.

    All projects generated are assumed to have the following context:

    * ``name`` - human-friendly name for the project, e.g. ``"Poser Plus"``
    * ``description`` - brief (one-line) description of the project
    * ``folder`` - folder name for the project, e.g. ``"poser-plus"``
    * ``pkg_name`` - package name for use in Python, e.g. ``"poser_plus"``
    * ``pypi_name`` - package name for use with PyPI, e.g. ``"Poser-Plus"``
    * ``egg_name`` - package name used with egg files, e.g. ``"Poser_Plus"``
    * ``studly_prefix`` - prefix for class names, e.g. ``PoserPlus``
    * ``env_name`` - name of the Python virtual environment
    * ``requires`` - dict of required dependencies
    * ``classifiers`` - set of Python trove classifiers for project
    * ``entry_points`` - dict of setuptools entry points for project
    """
    # nb. subclass must override this!
    key = 'python'

    def make_schema(self, **kwargs):
        schema = super(PythonProjectGenerator, self).make_schema(**kwargs)

        schema.add(colander.SchemaNode(name='name',
                                       typ=colander.String()))

        schema.add(colander.SchemaNode(name='pkg_name',
                                       typ=colander.String()))

        schema.add(colander.SchemaNode(name='pypi_name',
                                       typ=colander.String()))

        return schema

    def normalize_context(self, context):
        context = super(PythonProjectGenerator, self).normalize_context(context)

        if 'description' not in context:
            context['description'] = ""

        if 'folder' not in context:
            context['folder'] = context['pkg_name'].replace('_', '-')

        if 'egg_name' not in context:
            context['egg_name'] = context['pypi_name'].replace('-', '_')

        if 'studly_prefix' not in context:
            context['studly_prefix'] = get_studly_prefix(context['name'])

        if 'env_name' not in context:
            context['env_name'] = context['folder']

        if 'requires' not in context:
            context['requires'] = {}

        if 'classifiers' not in context:
            context['classifiers'] = set([
                'Development Status :: 3 - Alpha',
                'Intended Audience :: Developers',
                'Natural Language :: English',
                'Programming Language :: Python',
                'Programming Language :: Python :: 3',
            ])

        if 'entry_points' not in context:
            context['entry_points'] = {}

        return context

    def generate_project(self, output, context, **kwargs):

        ##############################
        # root project dir
        ##############################

        self.generate('gitignore.mako',
                      os.path.join(output, '.gitignore'),
                      context)

        self.generate('MANIFEST.in.mako',
                      os.path.join(output, 'MANIFEST.in'),
                      context)

        self.generate('README.md.mako',
                      os.path.join(output, 'README.md'),
                      context)

        self.generate('CHANGELOG.md.mako',
                      os.path.join(output, 'CHANGELOG.md'),
                      context)

        self.generate('setup.cfg.mako',
                      os.path.join(output, 'setup.cfg'),
                      context)

        self.generate('setup.py.mako',
                      os.path.join(output, 'setup.py'),
                      context)

        self.generate('tasks.py.mako',
                      os.path.join(output, 'tasks.py'),
                      context)

        ##############################
        # root package dir
        ##############################

        package = os.path.join(output, context['pkg_name'])
        os.makedirs(package)

        self.generate('package/__init__.py.mako',
                      os.path.join(package, '__init__.py'),
                      context)

        self.generate('package/_version.py',
                      os.path.join(package, '_version.py'))


class RattailAdjacentProjectGenerator(PythonProjectGenerator):
    """
    Base class for "Rattail-adjacent" project generators, i.e. for
    projects which are based on Rattail, but may or may not be full
    "Poser" type apps.

    In addition to normal context for Python projects, all
    Rattail-adjacent projects are assumed to have the following
    context:

    * ``extends_config`` - whether the app extends Rattail config
    * ``has_cli`` - whether the app has its own command interface
    * ``extends_db`` - whether the app extends DB schema
    * ``has_model`` - whether the app provides top-level ORM model
    """
    # nb. subclass must override this!
    key = 'rattail_adjacent'

    def make_schema(self, **kwargs):
        schema = super(RattailAdjacentProjectGenerator, self).make_schema(**kwargs)

        schema.add(colander.SchemaNode(name='extends_config',
                                       typ=colander.Boolean()))

        schema.add(colander.SchemaNode(name='has_cli',
                                       typ=colander.Boolean()))

        schema.add(colander.SchemaNode(name='extends_db',
                                       typ=colander.Boolean()))

        schema.add(colander.SchemaNode(name='has_model',
                                       typ=colander.Boolean()))

        return schema

    def normalize_context(self, context):
        context = super(RattailAdjacentProjectGenerator, self).normalize_context(context)

        context['requires'].setdefault('rattail', True)

        context['classifiers'].update(set([
            'Topic :: Office/Business',
        ]))

        if context['extends_config']:
            context['entry_points'].setdefault('rattail.config.extensions', []).extend([
                "{0} = {0}.config:{1}Config".format(context['pkg_name'],
                                                    context['studly_prefix'])])

        if context['has_cli']:
            context['entry_points'].setdefault('console_scripts', []).extend([
                "{0} = {0}.commands:main".format(context['pkg_name']),
            ])
            context['entry_points'].setdefault('{}.commands'.format(context['pkg_name']), []).extend([
                "hello = {}.commands:HelloWorld".format(context['pkg_name']),
                "install = {}.commands:Install".format(context['pkg_name']),
            ])

            # nb. these alembic values are only needed for installer
            # template, which means only used if `has_cli=True`

            if 'alembic_script_location' not in context:
                context['alembic_script_location'] = 'rattail.db:alembic'

            if 'alembic_version_locations' not in context:
                context['alembic_version_locations'] = ['rattail.db:alembic/versions']
            if context['extends_db']:
                context['alembic_version_locations'].append(
                    '{}.db:alembic/versions'.format(context['pkg_name']))

        return context

    def generate_project(self, output, context, **kwargs):
        from alembic.config import Config as AlembicConfig
        from alembic.command import revision as alembic_revision

        super(RattailAdjacentProjectGenerator, self).generate_project(
            output, context, **kwargs)

        ##############################
        # root package dir
        ##############################

        package = os.path.join(output, context['pkg_name'])

        if context['extends_config']:
            self.generate('package/config.py.mako',
                          os.path.join(package, 'config.py'),
                          context)

        if context['has_cli']:
            self.generate('package/commands.py.mako',
                          os.path.join(package, 'commands.py'),
                          context)

        ##############################
        # db package dir
        ##############################

        if context['has_model'] or context['extends_db']:

            db = os.path.join(package, 'db')
            os.makedirs(db)

            self.generate('package/db/__init__.py',
                          os.path.join(db, '__init__.py'))

            ####################
            # model
            ####################

            model = os.path.join(db, 'model')
            os.makedirs(model)

            self.generate('package/db/model/__init__.py.mako',
                          os.path.join(model, '__init__.py'),
                          context)

            ####################
            # alembic
            ####################

            if context['extends_db']:

                alembic = os.path.join(db, 'alembic')
                os.makedirs(alembic)

                versions = os.path.join(alembic, 'versions')
                os.makedirs(versions)

                # make alembic config, aware of new project versions folder
                alembic_config = AlembicConfig()
                alembic_config.set_main_option('script_location',
                                            'rattail.db:alembic')
                alembic_config.set_main_option('version_locations',
                                            '{} rattail.db:alembic/versions'.format(
                                                versions))

                # generate first revision script for new project
                script = alembic_revision(alembic_config,
                                          version_path=versions,
                                          head='rattail@head',
                                          splice=True,
                                          branch_label=context['pkg_name'],
                                          message="add {} branch".format(context['pkg_name']))

                # declare `down_revision = None` ..no way to tell alembic
                # to do that apparently, so we must rewrite file
                with open(script.path, 'rt') as f:
                    old_contents = f.read()
                new_contents = []
                for line in old_contents.split('\n'):
                    if line.startswith('down_revision ='):
                        line = re.sub(r"'\w+'", 'None', line)
                    new_contents.append(line)
                with open(script.path, 'wt') as f:
                    f.write('\n'.join(new_contents))

        ##############################
        # templates
        ##############################

        if context['has_cli']:

            templates = os.path.join(package, 'templates')
            os.makedirs(templates)

            installer = os.path.join(templates, 'installer')
            os.makedirs(installer)

            self.generate('package/templates/installer/rattail.conf.mako',
                          os.path.join(installer, 'rattail.conf.mako'),
                          context)

            self.generate('package/templates/installer/upgrade.sh.mako_',
                          os.path.join(installer, 'upgrade.sh.mako'))


class PoserProjectGenerator(RattailAdjacentProjectGenerator):
    """
    Base class for Poser project generators.

    In addition to normal context for "Rattail-adjacent" projects, all
    Poser projects are assumed to have the following context:

    * ``organization`` - human-friendly name for the organization
    * ``has_db`` - whether the app has a Rattail DB
    * ``has_batch_schema`` - whether the DB needs dynamic 'batch' schema
    * ``db_name`` - name of the Rattail DB
    * ``has_datasync`` - whether the app needs a datasync service
    * ``has_web`` - whether the app has tailbone web UI
    * ``has_web_api`` - whether the app has tailbone web API
    * ``beaker_session_secret`` - secret for Beaker session storage
    * ``uses_fabric`` - whether the app is deployed via fabric
    """
    # nb. subclass must override this!
    key = 'poser'

    def make_schema(self, **kwargs):
        schema = super(PoserProjectGenerator, self).make_schema(**kwargs)

        schema.add(colander.SchemaNode(name='organization',
                                       typ=colander.String()))

        schema.add(colander.SchemaNode(name='has_db',
                                       typ=colander.Boolean()))

        schema.add(colander.SchemaNode(name='has_batch_schema',
                                       typ=colander.Boolean()))

        schema.add(colander.SchemaNode(name='has_web',
                                       typ=colander.Boolean()))

        schema.add(colander.SchemaNode(name='has_web_api',
                                       typ=colander.Boolean()))

        schema.add(colander.SchemaNode(name='has_datasync',
                                       typ=colander.Boolean()))

        schema.add(colander.SchemaNode(name='uses_fabric',
                                       typ=colander.Boolean()))

        return schema

    def normalize_context(self, context):
        context = super(PoserProjectGenerator, self).normalize_context(context)

        if not context.get('description'):
            context['description'] = "Rattail/Poser project for {}".format(
                context['organization'])

        context['classifiers'].update(set([
            'Environment :: Console',
            'Operating System :: POSIX :: Linux',
        ]))
        if context['has_web'] or context['has_web_api']:
            context['classifiers'].update(set([
                'Environment :: Web Environment',
                'Framework :: Pyramid',
            ]))

        if context['has_web']:
            context['entry_points'].setdefault('paste.app_factory', []).extend([
                "main = {}.web.app:main".format(context['pkg_name']),
            ])

        if 'db_name' not in context:
            context['db_name'] = context['pkg_name']

        if 'beaker_session_secret' not in context:
            context['beaker_session_secret'] = self.random_string()

        if context['has_db']:
            context['requires'].setdefault('psycopg2', True)

        if context['has_web']:
            context['requires'].setdefault('Tailbone', True)
        elif context['has_db']:
            context['requires'].setdefault('rattail', 'rattail[db]')

        if context['uses_fabric']:
            context['requires'].setdefault('rattail-fabric2', True)

        if 'mako_directories' not in context:
            context['mako_directories'] = [
                '{}.web:templates'.format(context['pkg_name']),
                'tailbone:templates',
            ]

        return context

    def generate_project(self, output, context, **kwargs):
        super(PoserProjectGenerator, self).generate_project(
            output, context, **kwargs)

        package = os.path.join(output, context['pkg_name'])

        ##############################
        # web package dir
        ##############################

        if context['has_web']:

            web = os.path.join(package, 'web')
            os.makedirs(web)

            self.generate('package/web/__init__.py',
                          os.path.join(web, '__init__.py'))

            self.generate('package/web/app.py.mako',
                          os.path.join(web, 'app.py'),
                          context)

            self.generate('package/web/menus.py.mako', os.path.join(web, 'menus.py'),
                          context)

            self.generate('package/web/subscribers.py.mako',
                          os.path.join(web, 'subscribers.py'),
                          context)

            static = os.path.join(web, 'static')
            os.makedirs(static)

            self.generate('package/web/static/__init__.py.mako',
                          os.path.join(static, '__init__.py'),
                          context)

            web_templates = os.path.join(web, 'templates')
            os.makedirs(web_templates)

            self.generate('package/web/templates/base_meta.mako_tmpl',
                          os.path.join(web_templates, 'base_meta.mako'),
                          context)

            views = os.path.join(web, 'views')
            os.makedirs(views)

            self.generate('package/web/views/__init__.py.mako',
                          os.path.join(views, '__init__.py'),
                          context)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Product-related commands
"""

import logging

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand, date_argument
from .typer import typer_get_runas_user


log = logging.getLogger(__name__)


@rattail_typer.command()
def update_costs(
        ctx: typer.Context,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Update (move future to current) costs for all products
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    user = typer_get_runas_user(ctx)
    do_update_costs(config, user, dry_run=dry_run, progress=progress)


class UpdateCosts(Subcommand):
    """
    Update (move future to current) costs for all products
    """
    name = 'update-costs'
    description = __doc__.strip()

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def run(self, args):
        user = self.get_runas_user()
        do_update_costs(self.config, user, dry_run=args.dry_run, progress=self.progress)


def do_update_costs(config, user, dry_run=False, progress=None):
    from rattail.db.continuum import versioning_manager
    from rattail.db.util import finalize_session

    app = config.get_app()
    model = app.model
    session = app.make_session()
    user = session.get(model.User, user.uuid)
    session.set_continuum_user(user)
    products_handler = app.get_products_handler()

    # TODO: even if this works, it seems heavy-handed...
    # (also it *doesn't* work if ran before setting continuum user)
    uow = versioning_manager.unit_of_work(session)
    transaction = uow.create_transaction(session)
    transaction.meta = {'comment': "make future costs become current"}

    now = app.make_utc()
    future_costs = session.query(model.ProductFutureCost)\
                          .filter(model.ProductFutureCost.starts <= now)\
                          .all()
    log.info("found %s future costs which should become current", len(future_costs))

    def move(future, i):
        products_handler.make_future_cost_current(future)

    app.progress_loop(move, future_costs, progress,
                      message="Making future costs become current")

    finalize_session(session, dry_run=dry_run)



# -*- coding: utf-8; -*-
"""add tender.cash

Revision ID: 9cc5b433fb1d
Revises: dd1a440a93f1
Create Date: 2023-10-01 18:43:13.587891

"""

# revision identifiers, used by Alembic.
revision = '9cc5b433fb1d'
down_revision = 'dd1a440a93f1'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # tender
    op.add_column('tender', sa.Column('is_cash', sa.Boolean(), nullable=True))
    op.add_column('tender', sa.Column('allow_cash_back', sa.Boolean(), nullable=True))
    op.add_column('tender_version', sa.Column('is_cash', sa.Boolean(), autoincrement=False, nullable=True))
    op.add_column('tender_version', sa.Column('allow_cash_back', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # tender
    op.drop_column('tender_version', 'allow_cash_back')
    op.drop_column('tender_version', 'is_cash')
    op.drop_column('tender', 'allow_cash_back')
    op.drop_column('tender', 'is_cash')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Console Commands
"""

from .base import (rattail_typer,
                   RattailCommand, Command,
                   RattailSubcommand, Subcommand,
                   date_argument, dict_argument, list_argument,
                   main)
from .importing import ImportSubcommand, ImportFromCSV, ImportFileSubcommand, ExportFileSubcommand
from .install import InstallSubcommand

# TODO: is this the best we can do, to register available commands?
from . import backup
from . import batch
from . import bouncer
from . import checkdb
from . import cleanup
from . import clonedb
from . import datasync
from . import date_organize
from . import filemon
from . import luigi
from . import mailmon
from . import make_appdir
from . import make_config
from . import make_user
from . import make_uuid
from . import mysql
from . import postfix
from . import problems
from . import products
from . import projects
from . import run_n_mail
from . import runsql
from . import settings
from . import telemetry
from . import upgrade
from . import versions

# discover more commands
from .typer import typer_eager_imports
typer_eager_imports(rattail_typer)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Development Commands
"""

import os
import re
from pathlib import Path

import typer
from typing_extensions import Annotated
from mako.template import Template

from rattail.commands.typer import typer_callback
from rattail.files import resource_path


rattail_dev_typer = typer.Typer(
    callback=typer_callback,
    help="Rattail - development commands"
)


@rattail_dev_typer.command()
def new_batch(
        ctx: typer.Context,
        model: Annotated[
            str,
            typer.Argument(help="Name of primary model for new batch, e.g. "
                           "'VendorCatalog' or 'PrintLabels'.")],
        output_dir: Annotated[
            Path,
            typer.Option('--output-dir', '-O',
                         help="Path to which generated files should be written; "
                         "defaults to current directory.")] = '.',
):
    """
    Generate some code for a new batch type.
    """
    model_title = re.sub(r'([a-z])([A-Z])', r'\g<1> \g<2>', model)
    table_name = model_title.lower().replace(' ', '_')
    context = {
        'model_name': model,
        'model_title': model_title,
        'table_name': table_name,
    }
    template_dir = resource_path('rattail:data/new-batch')
    for name in ('model', 'handler', 'webview'):
        template_path = os.path.join(template_dir, f'{name}.py')
        template = Template(filename=template_path)
        output_path = os.path.join(output_dir, f'{name}.py')
        with open(output_path, 'wt') as output_file:
            output_file.write(template.render(**context))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Console Commands - base classes and core commands
"""

import os
import sys
import json
import argparse
import datetime
import warnings
import logging

from wuttjamaican.cmd.base import (Command as WuttaCommand,
                                   CommandArgumentParser,
                                   Subcommand as WuttaSubcommand)
from wuttjamaican.util import parse_list

from rattail import __version__
from rattail.progress import ConsoleProgress, SocketProgress
from rattail.config import make_config
from rattail.db.config import configure_versioning
from rattail.commands.typer import make_typer


rattail_typer = make_typer(
    name='rattail',
    help="Rattail Software Framework"
)


class ArgumentParser(CommandArgumentParser):
    """
    Custom argument parser.

    This is a compatibility wrapper around upstream
    :class:`wuttjamaican:wuttjamaican.commands.base.CommandArgumentParser`.
    New code should use that instead; this will eventually be removed.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn("the custom ArgumentParser in rattail is deprecated; "
                      "please use the one from wuttjamaican instead",
                      DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)


def date_argument(string):
    """
    Validate and coerce a date argument.

    This function is designed be used as the ``type`` parameter when calling
    ``ArgumentParser.add_argument()``, e.g.::

       parser = ArgumentParser()
       parser.add_argument('--date', type=date_argument)
    """
    try:
        date = datetime.datetime.strptime(string, '%Y-%m-%d').date()
    except ValueError:
        raise argparse.ArgumentTypeError("Date must be in YYYY-MM-DD format")
    return date


def dict_argument(string):
    """
    Coerce the given string to a Python dictionary.  The string is assumed to
    be JSON-encoded; this function merely invokes ``json.loads()`` on it.

    This function is designed be used as the ``type`` parameter when calling
    ``ArgumentParser.add_argument()``, e.g.::

       parser = ArgumentParser()
       parser.add_argument('--date', type=dict_argument)
    """
    try:
        return json.loads(string)
    except json.decoder.JSONDecodeError:
        raise argparse.ArgumentTypeError("Argument must be valid JSON-encoded string")


def list_argument(string):
    """
    Coerce the given string to a list of strings, splitting on whitespace
    and/or commas.

    This function is designed be used as the ``type`` parameter when calling
    ``ArgumentParser.add_argument()``, e.g.::

       parser = ArgumentParser()
       parser.add_argument('--things', type=list_argument)
    """
    return parse_list(string)


class RattailCommand(WuttaCommand):
    """
    The primary command for Rattail.

    This inherits from
    :class:`wuttjamaican:wuttjamaican.commands.base.Command` so see
    those docs for more info.

    Custom apps based on Rattail will probably want to make and
    register their own ``Command`` class derived from this one.  Again
    see upstream docs for more details.

    Rattail extends the upstream class by adding the following:
    """
    name = 'rattail'
    version = __version__
    description = "Rattail Software Framework"

    @property
    def db_config_section(self):
        """
        Name of section in config file which should have database connection
        info.  This defaults to ``'rattail.db'`` but may be overridden so the
        command framework can sit in front of a non-Rattail database if needed.

        This is used to auto-configure a "default" database engine for the app,
        when any command is invoked.
        """
        # TODO: surely this can be more dynamic? or is it really being used?
        return 'rattail.db'

    @property
    def db_session_factory(self):
        """
        Returns a reference to the configured session factory.

        This is a compatibility wrapper around
        :meth:`rattail.app.AppHandler.make_session()`.  New code
        should use that instead; this may eventually be removed.
        """
        return self.config.get_app().make_session

    @property
    def db_model(self):
        """
        Returns a reference to configured model module.

        This is a compatibility wrapper around
        :meth:`rattail.config.RattailConfig.get_model()`.  New
        code should use that instead; this may eventually be removed.
        """
        return self.config.get_model()

    def iter_subcommands(self):
        """
        Returns a generator for the subcommands, sorted by name.

        This should probably not be used; instead see upstream
        :meth:`wuttjamaican:wuttjamaican.commands.base.Command.sorted_subcommands()`.
        """
        for subcmd in self.sorted_subcommands():
            yield subcmd

    def add_args(self):
        """
        Configure args for the main command arg parser.

        Rattail extends the upstream
        :meth:`~wuttjamaican:wuttjamaican.commands.base.Command.add_args()`
        by adding various command line args which have traditionally
        been available for it.  Some of these may disappear some day
        but no changes are planned just yet.
        """
        super().add_args()
        parser = self.parser

        # TODO: i think these aren't really being used in practice..?
        parser.add_argument('-n', '--no-init', action='store_true', default=False)
        parser.add_argument('--no-extend-config', dest='extend_config', action='store_false')

        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('--progress-socket',
                            help="Optional socket (e.g. localhost:8487) to which progress info should be written.")
        parser.add_argument('--runas', '-R', metavar='USERNAME',
                            help="Optional username to impersonate when running the command.  "
                            "This is only relevant for / used by certain commands.")

        # data versioning
        parser.add_argument('--versioning', action='store_true',
                            help="Force *enable* of data versioning.  If set, then --no-versioning "
                            "cannot also be set.  If neither is set, config will determine whether "
                            "or not data versioning should be enabled.")
        parser.add_argument('--no-versioning', action='store_true',
                            help="Force *disable* of data versioning.  If set, then --versioning "
                            "cannot also be set.  If neither is set, config will determine whether "
                            "or not data versioning should be enabled.")

    def make_config(self, args):
        """
        Make the config object in preparation for running a subcommand.

        See also upstream
        :meth:`~wuttjamaican:wuttjamaican.commands.base.Command.make_config()`
        but for now, Rattail overrides this completely, mostly for the
        sake of versioning setup.
        """
        # TODO: can we make better args so this is handled by argparse somehow?
        if args.versioning and args.no_versioning:
            self.stderr.write("Cannot pass both --versioning and --no-versioning\n")
            sys.exit(1)

        # if args say not to "init" then we make a sort of empty config
        if args.no_init:
            config = make_config([], extend=False, versioning=False)

        else: # otherwise we make a proper config, and maybe turn on versioning
            logging.basicConfig()
            config = make_config(args.config_paths, plus_files=args.plus_config_paths,
                                 extend=args.extend_config, versioning=False)
            if args.versioning:
                configure_versioning(config, force=True)
            elif not args.no_versioning:
                configure_versioning(config)

        # import our primary data model now, just in case it hasn't fully been
        # imported yet.  this it to be sure association proxies and the like
        # are fully wired up in the case of extensions
        # TODO: what about web apps etc.? i guess i was only having the problem
        # for some importers, e.g. basic CORE API -> Rattail w/ the schema
        # extensions in place from rattail-corepos
        try:
            config.get_model()
        except ImportError:
            pass

        return config

    def prep_subcommand(self, subcmd, args):
        """
        Rattail overrides this method to apply some of the global args
        directly to the subcommand object.

        See also upstream
        :meth:`~wuttjamaican:wuttjamaican.commands.base.Command.prep_subcommand()`.
        """
        # figure out if/how subcommand should show progress
        subcmd.show_progress = args.progress
        subcmd.progress = None
        if subcmd.show_progress:
            if args.progress_socket:
                host, port = args.progress_socket.split(':')
                subcmd.progress = SocketProgress(host, int(port))
            else:
                subcmd.progress = ConsoleProgress

        # maybe should be verbose
        subcmd.verbose = args.verbose

        # TODO: make this default to something from config?
        subcmd.runas_username = args.runas or None


# TODO: deprecate / remove this?
Command = RattailCommand


class RattailSubcommand(WuttaSubcommand):
    """
    Base class for subcommands.

    This inherits from :class:`wuttjamaican.commands.base.Subcommand`
    so see those docs for more info.

    Rattail extends the subcommand to include:

    .. attribute:: runas_username

       Username (:attr:`~rattail.db.model.users.User.username`)
       corresponding to the :class:`~rattail.db.model.users.User`
       which the command should "run as" - i.e.  for sake of version
       history etc.

    .. attribute:: show_progress

       Boolean indicating whether progress should be shown for the
       subcommand.

    .. attribute:: progress

       Optional factory to be used when creating progress objects.
       This is ``None`` by default but if :attr:`show_progress` is
       enabled, then :class:`~rattail.progress.ConsoleProgress` is the
       default factory.

    .. attribute:: verbose

       Flag indicating the subcommand should be free to print
       arbitrary messages to
       :attr:`~wuttjamaican:wuttjamaican.commands.base.Subcommand.stdout`.
    """
    runas_username = None
    show_progress = False
    progress = None
    verbose = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # TODO: deprecate / remove this
        self.parent = self.command

    def add_args(self):
        """
        This is the "same" as upstream
        :meth:`wuttjamaican:wuttjamaican.commands.base.Subcommand.add_args()`
        except Rattail must customize this to also invoke its
        deprecated method, :meth:`add_parser_args()`.
        """
        super().add_args()
        self.add_parser_args(self.parser)

    def add_parser_args(self, parser):
        """
        This method is deprecated and will eventually be removed;
        please define :meth:`add_args()` instead.
        """

    @property
    def model(self):
        return self.parent.db_model

    def make_session(self):
        session = self.parent.db_session_factory()
        user = self.get_runas_user(session=session)
        if user:
            session.set_continuum_user(user)
        return session

    def finalize_session(self, *args, **kwargs):
        """
        Wrap up the given session, per the given arguments.  This is meant to
        provide a simple convenience, for commands which must do work within a
        DB session, but would like to support a "dry run" mode.
        """
        from rattail.db.util import finalize_session
        return finalize_session(*args, **kwargs)

    def get_runas_user(self, session=None, username=None):
        """
        Convenience method to get the "runas" User object for the
        current command.

        Uses :meth:`rattail.app.AppHandler.get_runas_user()` under the
        hood, but the ``--runas`` command line param provides the default
        username.
        """
        if not username:
            username = getattr(self, 'runas_username', None)
        return self.app.get_runas_user(session=session, username=username)

    def progress_loop(self, func, items, factory=None, **kwargs):
        return self.app.progress_loop(func, items, factory or self.progress, **kwargs)

    # TODO: deprecate / remove this
    def get_pip_path(self):
        return os.path.join(sys.prefix, 'bin', 'pip')

    def require_prompt_toolkit(self):
        from rattail.commands.util import require_prompt_toolkit
        return require_prompt_toolkit(self.config)

    def require_rich(self):
        from rattail.commands.util import require_rich
        return require_rich(self.config)

    def rprint(self, *args, **kwargs):
        from rattail.commands.util import rprint
        return rprint(*args, **kwargs)

    def basic_prompt(self, *args, **kwargs):
        from rattail.commands.util import basic_prompt
        return basic_prompt(*args, **kwargs)


# TODO: deprecate / remove this?
Subcommand = RattailSubcommand


def main(*args):
    """
    The primary entry point for the Rattail command system.
    """
    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Generator for 'tailbone-shopfoo' integration projects
"""

import os

from rattail.projects.tailbone_integration import TailboneIntegrationProjectGenerator


class TailboneShopfooProjectGenerator(TailboneIntegrationProjectGenerator):
    """
    Generator for projects which integrate Tailbone with some type of
    e-commerce system.  This is for generating projects such as
    tailbone-instacart, tailbone-mercato etc. which involve a nightly
    export/upload of product data to external server.
    """
    key = 'tailbone_shopfoo'

    def normalize_context(self, context):

        # nb. auto-set some flags
        context['has_web'] = False

        # then do normal logic
        context = super(TailboneShopfooProjectGenerator, self).normalize_context(context)

        return context

    def generate_project(self, output, context, **kwargs):
        super(TailboneShopfooProjectGenerator, self).generate_project(
            output, context, **kwargs)

        package = os.path.join(output, context['pkg_name'])

        ##############################
        # web views
        ##############################

        views = os.path.join(package, 'views')

        shopfoo = os.path.join(views, context['integration_pkgname'])
        os.makedirs(shopfoo)

        self.generate('package/views/shopfoo/__init__.py.mako',
                      os.path.join(shopfoo, '__init__.py'),
                      context)

        self.generate('package/views/shopfoo/products.py.mako',
                      os.path.join(shopfoo, 'products.py'),
                      context)

        self.generate('package/views/shopfoo/exports.py.mako',
                      os.path.join(shopfoo, 'exports.py'),
                      context)



# -*- coding: utf-8; -*-
"""rename batch_labels

Revision ID: 6a691bede834
Revises: 117236288811
Create Date: 2021-02-02 14:27:13.539293

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '6a691bede834'
down_revision = '117236288811'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # remove references to `label_batch`
    op.drop_constraint('label_batch_handheld_fk_batch', 'label_batch_handheld', type_='foreignkey')
    op.drop_constraint('label_batch_row_fk_batch', 'label_batch_row', type_='foreignkey')

    # batch_labels
    op.drop_constraint('label_batch_fk_created_by', 'label_batch', type_='foreignkey')
    op.drop_constraint('label_batch_fk_cognized_by', 'label_batch', type_='foreignkey')
    op.drop_constraint('label_batch_fk_executed_by', 'label_batch', type_='foreignkey')
    op.drop_constraint('label_batch_fk_handheld_batch', 'label_batch', type_='foreignkey')
    op.drop_constraint('label_batch_fk_label_profile', 'label_batch', type_='foreignkey')
    op.rename_table('label_batch', 'batch_labels')
    op.create_foreign_key('batch_labels_fk_created_by',
                          'batch_labels', 'user',
                          ['created_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_labels_fk_cognized_by',
                          'batch_labels', 'user',
                          ['cognized_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_labels_fk_executed_by',
                          'batch_labels', 'user',
                          ['executed_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_labels_fk_handheld_batch',
                          'batch_labels', 'batch_handheld',
                          ['handheld_batch_uuid'], ['uuid'])
    op.create_foreign_key('batch_labels_fk_label_profile',
                          'batch_labels', 'label_profile',
                          ['label_profile_uuid'], ['uuid'])

    # batch_labels_handheld
    op.drop_constraint('label_batch_handheld_fk_handheld', 'label_batch_handheld', type_='foreignkey')
    op.rename_table('label_batch_handheld', 'batch_labels_handheld')
    op.create_foreign_key('batch_labels_handheld_fk_batch',
                          'batch_labels_handheld', 'batch_labels',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('batch_labels_handheld_fk_handheld',
                          'batch_labels_handheld', 'batch_handheld',
                          ['handheld_uuid'], ['uuid'])

    # batch_labels_row
    op.drop_constraint('label_batch_row_fk_product', 'label_batch_row', type_='foreignkey')
    op.drop_constraint('label_batch_row_fk_label_profile', 'label_batch_row', type_='foreignkey')
    op.rename_table('label_batch_row', 'batch_labels_row')
    op.create_foreign_key('batch_labels_row_fk_batch_uuid',
                          'batch_labels_row', 'batch_labels',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('batch_labels_row_fk_product',
                          'batch_labels_row', 'product',
                          ['product_uuid'], ['uuid'])
    op.create_foreign_key('batch_labels_row_fk_label_profile',
                          'batch_labels_row', 'label_profile',
                          ['label_profile_uuid'], ['uuid'])


def downgrade():

    # remove references to `batch_labels`
    op.drop_constraint('batch_labels_handheld_fk_batch', 'batch_labels_handheld', type_='foreignkey')
    op.drop_constraint('batch_labels_row_fk_batch_uuid', 'batch_labels_row', type_='foreignkey')

    # batch_labels
    op.drop_constraint('batch_labels_fk_created_by', 'batch_labels', type_='foreignkey')
    op.drop_constraint('batch_labels_fk_cognized_by', 'batch_labels', type_='foreignkey')
    op.drop_constraint('batch_labels_fk_executed_by', 'batch_labels', type_='foreignkey')
    op.drop_constraint('batch_labels_fk_handheld_batch', 'batch_labels', type_='foreignkey')
    op.drop_constraint('batch_labels_fk_label_profile', 'batch_labels', type_='foreignkey')
    op.rename_table('batch_labels', 'label_batch')
    op.create_foreign_key('label_batch_fk_created_by',
                          'label_batch', 'user',
                          ['created_by_uuid'], ['uuid'])
    op.create_foreign_key('label_batch_fk_cognized_by',
                          'label_batch', 'user',
                          ['cognized_by_uuid'], ['uuid'])
    op.create_foreign_key('label_batch_fk_executed_by',
                          'label_batch', 'user',
                          ['executed_by_uuid'], ['uuid'])
    op.create_foreign_key('label_batch_fk_handheld_batch',
                          'label_batch', 'batch_handheld',
                          ['handheld_batch_uuid'], ['uuid'])
    op.create_foreign_key('label_batch_fk_label_profile',
                          'label_batch', 'label_profile',
                          ['label_profile_uuid'], ['uuid'])

    # batch_labels_handheld
    op.drop_constraint('batch_labels_handheld_fk_handheld', 'batch_labels_handheld', type_='foreignkey')
    op.rename_table('batch_labels_handheld', 'label_batch_handheld')
    op.create_foreign_key('label_batch_handheld_fk_batch',
                          'label_batch_handheld', 'label_batch',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('label_batch_handheld_fk_handheld',
                          'label_batch_handheld', 'batch_handheld',
                          ['handheld_uuid'], ['uuid'])

    # batch_labels_row
    op.drop_constraint('batch_labels_row_fk_product', 'batch_labels_row', type_='foreignkey')
    op.drop_constraint('batch_labels_row_fk_label_profile', 'batch_labels_row', type_='foreignkey')
    op.rename_table('batch_labels_row', 'label_batch_row')
    op.create_foreign_key('label_batch_row_fk_batch',
                          'label_batch_row', 'label_batch',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('label_batch_row_fk_product',
                          'label_batch_row', 'product',
                          ['product_uuid'], ['uuid'])
    op.create_foreign_key('label_batch_row_fk_label_profile',
                          'label_batch_row', 'label_profile',
                          ['label_profile_uuid'], ['uuid'])



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
``rattail upgrade`` command
"""

import logging
import sys

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand
from .typer import typer_get_runas_user


log = logging.getLogger(__name__)


@rattail_typer.command()
def upgrade(
        ctx: typer.Context,
        system: Annotated[
            str,
            typer.Option(help="System to which the upgrade applies.")] = 'rattail',
        description: Annotated[
            str,
            typer.Option(help="Description for the new/matched upgrade.")] = None,
        enabled: Annotated[
            bool,
            typer.Option(help="Desired enabled flag for the new/matched upgrade.")] = True,
        create: Annotated[
            bool,
            typer.Option('--create',
                         help="Create a new upgrade with the given attributes.")] = False,
        execute: Annotated[
            bool,
            typer.Option('--execute',
                         help="Execute the upgrade.  Note that if you do not specify "
                         "--create then the upgrade matching the given attributes "
                         "will be read from the database.  If such an upgrade is not "
                         "found or is otherwise invalid (e.g. already executed), "
                         "the command will fail.")] = False,
        keep_exit_code: Annotated[
            bool,
            typer.Option('--keep-exit-code',
                         help="Exit with same return code as subprocess.  If "
                         "this is not specified, this command will normally "
                         "exit with code 0 regardless of what happens with "
                         "the subprocess.  (only applicable with --execute)")] = False,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Upgrade the local Rattail app
    """
    from sqlalchemy import orm
    from rattail.db.util import finalize_session

    if not create and not execute:
        sys.stderr.write("Must specify --create and/or --execute\n")
        sys.exit(1)

    config = ctx.parent.rattail_config
    app = config.get_app()
    progress = ctx.parent.rattail_progress

    session = app.make_session()
    model = app.model
    user = typer_get_runas_user(ctx, session=session)

    if create:
        upgrade = model.Upgrade()
        upgrade.system = system or 'rattail'
        upgrade.description = description
        upgrade.created = app.make_utc()
        upgrade.created_by = user
        upgrade.enabled = enabled
        session.add(upgrade)
        session.flush()
        log.info("user '%s' created new upgrade: %s", user.username, upgrade)

    else:
        upgrades = session.query(model.Upgrade)\
                          .filter(model.Upgrade.enabled == enabled)
        if description:
            upgrades = upgrades.filter(model.Upgrade.description == description)
        try:
            upgrade = upgrades.one()
        except orm.exc.NoResultFound:
            sys.stderr.write("no matching upgrade found\n")
            session.rollback()
            session.close()
            sys.exit(1)
        except orm.exc.MultipleResultsFound:
            sys.stderr.write("found {} matching upgrades\n".format(upgrades.count()))
            session.rollback()
            session.close()
            sys.exit(1)

    if execute:
        if upgrade.executed:
            sys.stderr.write("upgrade has already been executed: {}\n".format(upgrade))
            session.rollback()
            session.close()
            sys.exit(1)
        if not upgrade.enabled:
            sys.stderr.write("upgrade is not enabled for execution: {}\n".format(upgrade))
            session.rollback()
            session.close()
            sys.exit(1)

        # execute upgrade
        handler = app.get_upgrade_handler()
        log.info("will now execute upgrade: %s", upgrade)
        if not dry_run:
            handler.mark_executing(upgrade)
            session.commit()
            handler.do_execute(upgrade, user, progress=progress)
        log.info("user '%s' executed upgrade: %s", user.username, upgrade)

    finalize_session(session, dry_run=dry_run)

    if (execute and not dry_run
        and keep_exit_code and upgrade.exit_code):
        sys.exit(upgrade.exit_code)


class Upgrade(Subcommand):
    """
    Upgrade the local Rattail app
    """
    name = 'upgrade'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('--system', default='rattail',
                            help="System to which the upgrade applies.")

        parser.add_argument('--description',
                            help="Description for the new/matched upgrade.")
        parser.add_argument('--enabled', action='store_true', default=True,
                            help="Indicate the enabled flag should be ON for the new/matched upgrade.  "
                            "Note that this is the default if you do not specify.")
        parser.add_argument('--no-enabled', action='store_false', dest='enabled',
                            help="Indicate the enabled flag should be OFF for the new/matched upgrade.")
        parser.add_argument('--create', action='store_true',
                            help="Create a new upgrade with the given attributes.")
        parser.add_argument('--execute', action='store_true',
                            help="Execute the upgrade.  Note that if you do not specify "
                            "--create then the upgrade matching the given attributes "
                            "will be read from the database.  If such an upgrade is not "
                            "found or is otherwise invalid (e.g. already executed), "
                            "the command will fail.")

        parser.add_argument('--keep-exit-code', action='store_true',
                            help="Exit with same return code as subprocess.  If "
                            "this is not specified, this command will normally "
                            "exit with code 0 regardless of what happens with "
                            "the subprocess.  (only applicable with --execute)")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def run(self, args):
        from sqlalchemy import orm

        if not args.create and not args.execute:
            self.stderr.write("Must specify --create and/or --execute\n")
            sys.exit(1)

        session = self.make_session()
        model = self.model
        user = self.get_runas_user(session)

        if args.create:
            upgrade = model.Upgrade()
            upgrade.system = args.system or 'rattail'
            upgrade.description = args.description
            upgrade.created = self.app.make_utc()
            upgrade.created_by = user
            upgrade.enabled = args.enabled
            session.add(upgrade)
            session.flush()
            log.info("user '%s' created new upgrade: %s", user.username, upgrade)

        else:
            upgrades = session.query(model.Upgrade)\
                              .filter(model.Upgrade.enabled == args.enabled)
            if args.description:
                upgrades = upgrades.filter(model.Upgrade.description == args.description)
            try:
                upgrade = upgrades.one()
            except orm.exc.NoResultFound:
                self.stderr.write("no matching upgrade found\n")
                session.rollback()
                session.close()
                sys.exit(1)
            except orm.exc.MultipleResultsFound:
                self.stderr.write("found {} matching upgrades\n".format(upgrades.count()))
                session.rollback()
                session.close()
                sys.exit(1)

        if args.execute:
            if upgrade.executed:
                self.stderr.write("upgrade has already been executed: {}\n".format(upgrade))
                session.rollback()
                session.close()
                sys.exit(1)
            if not upgrade.enabled:
                self.stderr.write("upgrade is not enabled for execution: {}\n".format(upgrade))
                session.rollback()
                session.close()
                sys.exit(1)

            # execute upgrade
            handler = self.app.get_upgrade_handler()
            log.info("will now execute upgrade: %s", upgrade)
            if not args.dry_run:
                handler.mark_executing(upgrade)
                session.commit()
                handler.do_execute(upgrade, user, progress=self.progress)
            log.info("user '%s' executed upgrade: %s", user.username, upgrade)

        if args.dry_run:
            session.rollback()
            log.info("dry run, so transaction was rolled back")
        else:
            session.commit()
            log.info("transaction was committed")
        session.close()

        if (args.execute and not args.dry_run
            and args.keep_exit_code and upgrade.exit_code):
            sys.exit(upgrade.exit_code)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Products API
"""

import warnings

from sqlalchemy import orm

from rattail import enum
from rattail.db import model


def get_product_by_upc(session, upc, include_deleted=False):
    """ DEPRECATED """
    warnings.warn("get_product_by_upc() is deprecated; please use "
                  "ProductHandler.locate_product_for_gpc() instead",
                  DeprecationWarning, stacklevel=2)

    products = session.query(model.Product)
    if not include_deleted:
        products = products.filter(model.Product.deleted == False)
    try:
        return products.filter(model.Product.upc == upc).one()
    except orm.exc.NoResultFound:
        pass


def get_product_by_item_id(session, item_id, include_deleted=False):
    """ DEPRECATED """
    warnings.warn("get_product_by_item_id() is deprecated; please use "
                  "ProductHandler.locate_product_for_item_id() instead",
                  DeprecationWarning, stacklevel=2)

    products = session.query(model.Product)
    if not include_deleted:
        products = products.filter(model.Product.deleted == False)
    try:
        return products.filter(model.Product.item_id == item_id).one()
    except orm.exc.NoResultFound:
        pass


def get_product_by_scancode(session, scancode, include_deleted=False):
    """ DEPRECATED """
    warnings.warn("get_product_by_scancode() is deprecated; please use "
                  "ProductHandler.locate_product_for_scancode() instead",
                  DeprecationWarning, stacklevel=2)

    products = session.query(model.Product)
    if not include_deleted:
        products = products.filter(model.Product.deleted == False)
    try:
        return products.filter(model.Product.scancode == scancode).one()
    except orm.exc.NoResultFound:
        pass


def get_product_by_code(session, code, include_deleted=False):
    """ DEPRECATED """
    warnings.warn("get_product_by_code() is deprecated; please use "
                  "ProductHandler.locate_product_for_alt_code() instead",
                  DeprecationWarning, stacklevel=2)

    products = session.query(model.Product).join(model.ProductCode)
    if not include_deleted:
        products = products.filter(model.Product.deleted == False)
    return products.filter(model.ProductCode.code == code).first()


def get_product_by_vendor_code(session, code, vendor=None, include_deleted=False):
    """ DEPRECATED """
    warnings.warn("get_product_by_vendor_code() is deprecated; please use "
                  "ProductHandler.locate_product_for_vendor_code() instead",
                  DeprecationWarning, stacklevel=2)

    products = session.query(model.Product).join(model.ProductCost)
    if vendor:
        products = products.filter(model.ProductCost.vendor == vendor)
    if not include_deleted:
        products = products.filter(model.Product.deleted == False)
    return products.filter(model.ProductCost.code == code).first()


def set_regular_price(product, newprice, **kwargs):
    """
    Simple way to set the regular price for the product, using sane defaults.
    """
    regular = product.regular_price
    if newprice is None:
        if regular:
            product.regular_price = None
        return

    if regular and regular.price == newprice:
        return                  # nothing to do

    price = regular or model.ProductPrice()
    price.type = kwargs.get('type', enum.PRICE_TYPE_REGULAR)
    price.level = kwargs.get('level', 1)
    price.price = newprice
    price.multiple = kwargs.get('multiple', 1)
    price.starts = kwargs.get('starts', None)
    price.ends = kwargs.get('ends', None)
    if price is not regular:
        product.prices.append(price)
    product.regular_price = price
    return price


def set_current_sale_price(product, newprice, starts=None, ends=None, **kwargs):
    """
    Simple way to set the current sale price for the product, using sane
    defaults.  Note that this does not check to confirm that your date range
    encompasses "today".
    """
    current = product.current_price
    if newprice is None:
        if current:
            product.current_price = None
        return
        
    if (current and current.price == newprice
        and current.starts == starts and current.ends == ends):
        return                  # nothing to do

    price = current or model.ProductPrice()
    price.type = kwargs.get('type', enum.PRICE_TYPE_SALE)
    price.level = kwargs.get('level', 1)
    price.price = newprice
    price.multiple = kwargs.get('multiple', 1)
    price.starts = starts
    price.ends = ends
    if price is not current:
        product.prices.append(price)
    product.current_price = price
    return price


future_cost_fieldmap = {
    'order_code': 'code',
    'case_quantity': 'case_size',
    'case_cost': 'case_cost',
    'unit_cost': 'unit_cost',
    'starts': 'effective',
    'discontinued': 'discontinued',
}


def make_future_cost_current(config, future):
    """ DEPRECATED """
    warnings.warn("get_product_by_upc() is deprecated; please use "
                  "ProductHandler.locate_product_for_gpc() instead",
                  DeprecationWarning, stacklevel=2)

    app = config.get_app()
    products_handler = app.get_products_handler()
    return products_handler.make_future_cost_current(future)



# -*- coding: utf-8; -*-
"""add equity_payment.status_code

Revision ID: 17cd825534de
Revises: 41a05b420280
Create Date: 2023-11-05 16:33:27.646431

"""

# revision identifiers, used by Alembic.
revision = '17cd825534de'
down_revision = '41a05b420280'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # member_equity_payment
    op.add_column('member_equity_payment', sa.Column('status_code', sa.Integer(), nullable=True))
    op.add_column('member_equity_payment_version', sa.Column('status_code', sa.Integer(), autoincrement=False, nullable=True))


def downgrade():

    # member_equity_payment
    op.drop_column('member_equity_payment_version', 'status_code')
    op.drop_column('member_equity_payment', 'status_code')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Utilities
"""

import decimal
import re
import pprint
import logging
import warnings

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.associationproxy import ASSOCIATION_PROXY

# TODO: Deprecate/remove these imports.
from rattail.db.config import get_engines, get_default_engine, configure_session


log = logging.getLogger(__name__)


class CounterMagic(object):
    """
    Provides magic counter values, to simulate PostgreSQL sequence.
    """

    def __init__(self, config):
        self.config = config
        self.metadata = sa.MetaData()

    def next_value(self, session, key):
        """
        Increment and return the next counter value for given key.
        """
        engine = session.bind
        table = sa.Table('counter_{}'.format(key), self.metadata,
                         sa.Column('value', sa.Integer(), primary_key=True))
        table.create(engine, checkfirst=True)
        with engine.begin() as cxn:
            result = cxn.execute(table.insert())
            return result.lastrowid


class QuerySequence(object):
    """
    Simple wrapper for a SQLAlchemy (or Django, or other?) query, to make it
    sort of behave like a normal sequence, as much as needed to e.g. make an
    importer happy.
    """

    def __init__(self, query):
        self.query = query

    def __len__(self):
        try:
            return len(self.query)
        except TypeError:
            return self.query.count()

    def __iter__(self):
        return iter(self.query)


def short_session(
        session=None,
        Session=None,
        commit=False,
        factory=None,
        config=None):
    """
    Compatibility wrapper around
    :class:`wuttjamaican:wuttjamaican.db.sess.short_session`.

    Note that this wrapper is a function whereas the upsream version
    is a proper context manager (class).  So calling this function
    will return a new instance of the upsream class.

    You should always specify keyword arguments when calling this
    function, since the arg order is different between this function
    and the upstream class.  And note that this function will
    eventually be deprecated and removed, so new code should call
    upstream directly.
    """
    from wuttjamaican.db import short_session

    warnings.warn("rattail.db.util.short_session() is deprecated; "
                  "please use wuttjamaican.db.short_session() instead",
                  DeprecationWarning, stacklevel=2)

    if not factory and Session:
        warnings.warn("passing a 'Session' kwarg is deprecated; "
                      "please pass 'factory' instead",
                      DeprecationWarning, stacklevel=2)
        factory = Session

    if not session and not factory and not config:
        from rattail.db import Session
        factory = Session

    return short_session(config=config, factory=factory, session=session, commit=commit)


def finalize_session(session, dry_run=False, success=True):
    """
    Wrap up the given session, per the given arguments.  This is meant
    to provide a simple convenience, for commands which must do work
    within a DB session, but would like to support a "dry run" mode.
    """
    if dry_run:
        session.rollback()
        log.info("dry run, so transaction was aborted")
    elif success:
        session.commit()
        log.info("transaction was committed")
    else:
        session.rollback()
        log.warning("action failed, so transaction was rolled back")
    session.close()


def get_fieldnames(config, obj, columns=True, proxies=True,
                   relations=False):
    """
    Produce a simple list of fieldnames for the given class,
    reflecting its table columns as well as any association proxies,
    and optionally, relationships.

    :param obj: Either a class or instance of a class, which derives
       from the base model class.

    :param columns: Whether or not to include simple columns.

    :param relations: Whether or not to include fields which represent
       relationships to other models.  If ``False`` (the default) then
       only "simple" fields will be included.

    :param proxies: Whether or not to include association proxy fields.
    """
    if isinstance(obj, type):
        cls = obj
    else:
        cls = obj.__class__

    mapper = orm.class_mapper(cls)
    fields = []

    # columns + relations
    prop_classes = []
    if columns:
        prop_classes.append(orm.ColumnProperty)
    if relations:
        prop_classes.append(orm.RelationshipProperty)
    if prop_classes:
        prop_classes = tuple(prop_classes)
        fields.extend([prop.key for prop in mapper.iterate_properties
                       if isinstance(prop, prop_classes)
                       and not prop.key.startswith('_')
                       and prop.key != 'versions'])

    # proxies
    if proxies:
        for key, desc in sa.inspect(cls).all_orm_descriptors.items():
            if desc.extension_type == ASSOCIATION_PROXY:

                # must avoid association proxies which in turn use
                # relationships, unless those are wanted by caller
                if not relations:
                    # TODO: this probably needs help, i stumbled thru it..
                    prop = sa.inspect(desc.for_class(cls).target_class)\
                             .get_property(desc.value_attr)
                    if isinstance(prop, orm.RelationshipProperty):
                        continue

                fields.append(key)

    return fields


def maxlen(attr):
    """
    Return the maximum length for the given attribute.
    """
    if len(attr.property.columns) == 1:
        type_ = attr.property.columns[0].type
        return getattr(type_, 'length', None)


def maxval(attr):
    """
    Return the maximum value possible for the given attribute.
    """
    if len(attr.property.columns) == 1:
        typ = attr.property.columns[0].type

        if isinstance(typ, sa.Numeric) and not isinstance(typ, sa.Float):
            maxint = pow(10, (typ.precision - typ.scale)) - 1
            return decimal.Decimal('{}.{}'.format(maxint,
                                                  '9' * typ.scale))


def make_topo_sortkey(model, metadata=None):
    """
    Returns a function suitable for use as a ``key`` kwarg to a standard Python
    sorting call.  This key function will expect a single class mapper and
    return a sequence number associated with that model.  The sequence is
    determined by SQLAlchemy's topological table sorting.
    """
    if metadata is None:
        metadata = model.Base.metadata

    tables = {}
    for i, table in enumerate(metadata.sorted_tables, 1):
        tables[table.name] = i

    # log.debug("topo sortkeys for '{}' will be:\n{}".format(model.__name__, pprint.pformat(
    #     [(i, name) for name, i in sorted(tables.items(), key=lambda t: t[1])])))

    def sortkey(name):
        if hasattr(model, name):
            mapper = orm.class_mapper(getattr(model, name))
            return tuple(tables[t.name] for t in mapper.tables)
        else:
            return tuple()

    return sortkey


def make_full_description(brand_name, description, size):
    """
    Combine the given field values into a complete description.
    """
    fields = [
        brand_name or '',
        description or '',
        size or '']
    fields = [f.strip() for f in fields if f.strip()]
    return ' '.join(fields)


def normalize_full_name(first_name, last_name):
    """
    Normalize the given first and last name to a "full" name value.  The
    fallback return value is an empty string.
    """
    first_name = (first_name or '').strip()
    last_name = (last_name or '').strip()
    if first_name and last_name:
        return "{} {}".format(first_name, last_name)
    if first_name:
        return first_name
    if last_name:
        return last_name
    return ''


##############################
# phone number validation
##############################

class PhoneValidator(object):
    """
    Simple validator, used to ensure a phone number matches the general
    expected pattern.
    """
    # NOTE: this was stolen from FormEncode
    _phoneRE = re.compile(r'^\s*(?:1-)?(\d\d\d)[\- \.]?(\d\d\d)[\- \.]?(\d\d\d\d)(?:\s*ext\.?\s*(\d+))?\s*$', re.I)

    def __init__(self, error=False):
        self.error = error

    def validate(self, number):
        if number:
            if not self._phoneRE.search(number):
                raise ValueError("Phone number is not valid")
            return number


def validate_phone_number(number, error=False):
    """
    Validate a single phone number.
    """
    validator = PhoneValidator(error=error)
    return validator.validate(number)


def normalize_phone_number(number):
    """
    Normalize a phone number to digits only.
    """
    if number is not None:
        return re.sub(r'\D', '', number)


def format_phone_number(number):
    """
    Returns a phone number in ``(XXX) XXX-XXXX`` format if possible; otherwise
    returns the argument unaltered.
    """
    original, number = number, normalize_phone_number(number)
    if number and len(number) == 10:
        return '({}) {}-{}'.format(number[:3], number[3:6], number[6:])
    return original



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
``rattail runsql`` command
"""

import argparse
import sys

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


@rattail_typer.command()
def runsql(
        ctx: typer.Context,
        engine: Annotated[
            str,
            typer.Argument(help="SQLAlchemy engine URL for the database.")] = ...,
        script: Annotated[
            typer.FileText,
            typer.Argument(help="File which contains a SQL script.")] = ...,
        max_width: Annotated[
            int,
            typer.Option(help="Max table width when displaying results.")] = 80,
):
    """
    Run (first statement of) a SQL script against a database
    """
    import sqlalchemy as sa
    import texttable

    sql = []
    for line in script:
        line = line.strip()
        if line and not line.startswith('--'):
            sql.append(line)
            if line.endswith(';'):
                break

    sql = ' '.join(sql)
    engine = sa.create_engine(engine)

    with engine.begin() as cxn:
        result = cxn.execute(sa.text(sql))
        rows = result.fetchall()
        if rows:
            table = texttable.Texttable(max_width=max_width)

            # force all columns to be treated as text.  that seems a bit
            # heavy-handed but is definitely the simplest way to deal with
            # issues such as a UPC being displayed in scientific notation
            table.set_cols_dtype(['t' for col in rows[0]])

            # add a header row, plus all data rows
            table.add_rows([rows[0].keys()] + rows)

            sys.stdout.write("{}\n".format(table.draw()))


class RunSQL(Subcommand):
    """
    Run (first statement of) a SQL script against a database
    """
    name = 'runsql'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('engine',
                            help="SQLAlchemy engine URL for the database.")
        parser.add_argument('script', type=argparse.FileType('r'),
                            help="Path to file which contains a SQL script.")
        parser.add_argument('--max-width', type=int, default=80,
                            help="Max table width when displaying results.")

    def run(self, args):
        import sqlalchemy as sa
        import texttable

        sql = []
        for line in args.script:
            line = line.strip()
            if line and not line.startswith('--'):
                sql.append(line)
                if line.endswith(';'):
                    break

        sql = ' '.join(sql)
        engine = sa.create_engine(args.engine)

        with engine.begin() as cxn:
            result = cxn.execute(sa.text(sql))
            rows = result.fetchall()
            if rows:
                table = texttable.Texttable(max_width=args.max_width)

                # force all columns to be treated as text.  that seems a bit
                # heavy-handed but is definitely the simplest way to deal with
                # issues such as a UPC being displayed in scientific notation
                table.set_cols_dtype(['t' for col in rows[0]])

                # add a header row, plus all data rows
                table.add_rows([rows[0].keys()] + rows)

                self.stdout.write("{}\n".format(table.draw()))



# -*- coding: utf-8; -*-

import configparser
import datetime
import os
from unittest import TestCase
from unittest.mock import patch

from wuttjamaican.testing import FileConfigTestCase

from rattail import config
from rattail.app import AppHandler


class TestParseBoolFunc(TestCase):

    def test_none(self):
        self.assertIsNone(config.parse_bool(None))

    def test_true(self):
        self.assertIs(config.parse_bool(True), True)

    def test_false(self):
        self.assertIs(config.parse_bool(False), False)

    def test_string(self):
        self.assertTrue(config.parse_bool('true'))
        self.assertTrue(config.parse_bool('yes'))
        self.assertTrue(config.parse_bool('on'))
        self.assertTrue(config.parse_bool('1'))

        self.assertFalse(config.parse_bool('false'))
        self.assertFalse(config.parse_bool('no'))
        self.assertFalse(config.parse_bool('off'))
        self.assertFalse(config.parse_bool('0'))


class TestParseListFunc(TestCase):

    def test_none(self):
        value = config.parse_list(None)
        self.assertEqual(len(value), 0)

    def test_single_value(self):
        value = config.parse_list(u'foo')
        self.assertEqual(len(value), 1)
        self.assertEqual(value[0], u'foo')

    def test_single_value_padded_by_spaces(self):
        value = config.parse_list(u'   foo   ')
        self.assertEqual(len(value), 1)
        self.assertEqual(value[0], u'foo')

    def test_slash_is_not_a_separator(self):
        value = config.parse_list(u'/dev/null')
        self.assertEqual(len(value), 1)
        self.assertEqual(value[0], u'/dev/null')

    def test_multiple_values_separated_by_whitespace(self):
        value = config.parse_list(u'foo bar baz')
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'bar')
        self.assertEqual(value[2], u'baz')

    def test_multiple_values_separated_by_commas(self):
        value = config.parse_list(u'foo,bar,baz')
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'bar')
        self.assertEqual(value[2], u'baz')

    def test_multiple_values_separated_by_whitespace_and_commas(self):
        value = config.parse_list(u'  foo,   bar   baz')
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'bar')
        self.assertEqual(value[2], u'baz')

    def test_multiple_values_separated_by_whitespace_and_commas_with_some_quoting(self):
        value = config.parse_list("""
        foo
        "C:\\some path\\with spaces\\and, a comma",
        baz
        """)
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'C:\\some path\\with spaces\\and, a comma')
        self.assertEqual(value[2], u'baz')

    def test_multiple_values_separated_by_whitespace_and_commas_with_single_quotes(self):
        value = config.parse_list("""
        foo
        'C:\\some path\\with spaces\\and, a comma',
        baz
        """)
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], 'foo')
        self.assertEqual(value[1], 'C:\\some path\\with spaces\\and, a comma')
        self.assertEqual(value[2], 'baz')


class TestRattailConfig(FileConfigTestCase):

    def setup_files(self):
        self.site_path = self.write_file('site.conf', """
[rattail]
        """)

        self.host_path = self.write_file('host.conf', """
[rattail.config]
include = "{}"
        """.format(self.site_path))

        self.app_path = self.write_file('app.conf', """
[rattail.config]
include = "{}"
        """.format(self.host_path))

        self.custom_path = self.write_file('custom.conf', """
[rattail.config]
include = "%(here)s/app.conf"
        """)

    def test_init_defaults(self):
        cfg = config.RattailConfig()
        self.assertEqual(cfg.files_requested, [])
        self.assertEqual(cfg.files_read, [])

    def test_init_params(self):
        self.setup_files()

        # files
        cfg = config.RattailConfig()
        self.assertEqual(cfg.files_requested, [])
        self.assertEqual(cfg.files_read, [])
        cfg = config.RattailConfig(files=[self.site_path])
        self.assertEqual(cfg.files_requested, [self.site_path])
        self.assertEqual(cfg.files_read, [self.site_path])

        # usedb
        cfg = config.RattailConfig()
        self.assertFalse(cfg.usedb)
        cfg = config.RattailConfig(usedb=True)
        self.assertTrue(cfg.usedb)

        # preferdb
        cfg = config.RattailConfig()
        self.assertFalse(cfg.preferdb)
        cfg = config.RattailConfig(preferdb=True)
        self.assertTrue(cfg.preferdb)

    def test_read_file_with_recurse(self):
        self.setup_files()
        cfg = config.RattailConfig()
        cfg.read_file(self.custom_path, recurse=True)
        self.assertEqual(cfg.files_requested, [self.custom_path, self.app_path, self.host_path, self.site_path])
        self.assertEqual(cfg.files_read, [self.site_path, self.host_path, self.app_path, self.custom_path])

    def test_read_file_once_only(self):
        self.setup_files()

        another_path = self.write_file('another.conf', """
[rattail.config]
include = "{custom}" "{site}" "{app}" "{site}" "{custom}"
        """.format(custom=self.custom_path, app=self.app_path, site=self.site_path))

        cfg = config.RattailConfig()
        cfg.read_file(another_path, recurse=True)
        self.assertEqual(cfg.files_requested, [another_path, self.custom_path, self.app_path, self.host_path, self.site_path])
        self.assertEqual(cfg.files_read, [self.site_path, self.host_path, self.app_path, self.custom_path, another_path])

    def test_read_file_skip_missing(self):
        self.setup_files()
        bogus_path = '/tmp/does-not/exist'
        self.assertFalse(os.path.exists(bogus_path))

        another_path = self.write_file('another.conf', """
[rattail.config]
include = "{bogus}" "{app}" "{bogus}" "{site}"
        """.format(bogus=bogus_path, app=self.app_path, site=self.site_path))

        cfg = config.RattailConfig()
        cfg.read_file(another_path, recurse=True)
        self.assertEqual(cfg.files_requested, [another_path, bogus_path, self.app_path, self.host_path, self.site_path])
        self.assertEqual(cfg.files_read, [self.site_path, self.host_path, self.app_path, another_path])

    @patch('rattail.config.logging.config.fileConfig')
    def test_configure_logging(self, fileConfig):
        cfg = config.RattailConfig()

        # logging not configured by default
        cfg.configure_logging()
        self.assertFalse(fileConfig.called)

        # but config option can enable it
        cfg.set('rattail.config', 'configure_logging', 'true')
        cfg.configure_logging()
        self.assertEqual(fileConfig.call_count, 1)
        fileConfig.reset_mock()

        # invalid logging config is ignored
        fileConfig.side_effect = configparser.NoSectionError('loggers')
        cfg.configure_logging()
        self.assertEqual(fileConfig.call_count, 1)


class TestRattailConfig(FileConfigTestCase):

    def test_prioritized_files(self):
        first = self.write_file('first.conf', """\
[foo]
bar = 1
""")

        second = self.write_file('second.conf', """\
[rattail.config]
require = %(here)s/first.conf
""")

        myconfig = config.RattailConfig(files=[second])
        files = myconfig.prioritized_files
        self.assertEqual(len(files), 2)
        self.assertEqual(files[0], second)
        self.assertEqual(files[1], first)
        self.assertIs(files, myconfig.get_prioritized_files())

    def test_setdefault(self):
        myconfig = config.RattailConfig()

        # nb. the tests below are effectively testing the custom get()
        # method in addition to setdefault()

        # value is empty by default
        self.assertIsNone(myconfig.get('foo.bar'))
        self.assertIsNone(myconfig.get('foo', 'bar'))

        # but we can change that by setting default
        myconfig.setdefault('foo.bar', 'baz')
        self.assertEqual(myconfig.get('foo.bar'), 'baz')
        self.assertEqual(myconfig.get('foo', 'bar'), 'baz')

        # also can set a default via section, option (as well as key)
        self.assertIsNone(myconfig.get('foo.blarg'))
        myconfig.setdefault('foo' ,'blarg', 'blast')
        self.assertEqual(myconfig.get('foo.blarg'), 'blast')
        self.assertEqual(myconfig.get('foo', 'blarg'), 'blast')

        # error is raised if args are ambiguous
        self.assertRaises(ValueError, myconfig.setdefault, 'foo', 'bar', 'blarg', 'blast')

        # try that for get() too
        self.assertRaises(ValueError, myconfig.get, 'foo', 'bar', 'blarg', 'blast')

    def test_getbool(self):
        myconfig = config.RattailConfig()
        self.assertFalse(myconfig.getbool('foo.bar'))
        myconfig.setdefault('foo.bar', 'true')
        self.assertTrue(myconfig.getbool('foo.bar'))

    def test_getint(self):
        myconfig = config.RattailConfig()
        self.assertIsNone(myconfig.getint('foo.bar'))
        myconfig.setdefault('foo.bar', '42')
        self.assertEqual(myconfig.getint('foo.bar'), 42)

    def test_getlist(self):
        myconfig = config.RattailConfig()
        self.assertIsNone(myconfig.getlist('foo.bar'))
        myconfig.setdefault('foo.bar', 'hello world')
        self.assertEqual(myconfig.getlist('foo.bar'), ['hello', 'world'])

    def test_getdate(self):
        myconfig = config.RattailConfig()
        self.assertIsNone(myconfig.getdate('foo.date'))
        myconfig.setdefault('foo.date', '2023-11-20')
        value = myconfig.getdate('foo.date')
        self.assertIsInstance(value, datetime.date)
        self.assertEqual(value, datetime.date(2023, 11, 20))

    def test_get_app(self):
        myconfig = config.RattailConfig()
        app = myconfig.get_app()
        self.assertIsInstance(app, AppHandler)
        self.assertIs(type(app), AppHandler)

    def test_parse_bool(self):
        myconfig = config.RattailConfig()
        self.assertTrue(myconfig.parse_bool('true'))
        self.assertFalse(myconfig.parse_bool('false'))

    def test_parse_list(self):
        myconfig = config.RattailConfig()
        self.assertEqual(myconfig.parse_list(None), [])
        self.assertEqual(myconfig.parse_list('hello world'), ['hello', 'world'])

    def test_make_list_string(self):
        myconfig = config.RattailConfig()

        value = myconfig.make_list_string(['foo', 'bar'])
        self.assertEqual(value, 'foo, bar')

        value = myconfig.make_list_string(['hello world', 'how are you'])
        self.assertEqual(value, "'hello world', 'how are you'")

        value = myconfig.make_list_string(["you don't", 'say'])
        self.assertEqual(value, "\"you don't\", say")

    def test_beaker_invalidate_setting(self):
        # TODO: this doesn't really test anything, just gives coverage
        myconfig = config.RattailConfig()
        myconfig.beaker_invalidate_setting('foo')


class TestRattailDefaultFiles(FileConfigTestCase):

    def test_quiet_conf(self):
        generic = self.write_file('generic.conf', '')
        quiet = self.write_file('quiet.conf', '')

        with patch('rattail.config.generic_default_files') as generic_default_files:
            generic_default_files.return_value = [generic]

            with patch('rattail.config.os') as mockos:
                mockos.path.join.return_value = quiet

                # generic files by default
                mockos.path.exists.return_value = False
                files = config.rattail_default_files('rattail')
                generic_default_files.assert_called_once_with('rattail')
                self.assertEqual(files, [generic])

                # but if quiet.conf exists, will return that
                generic_default_files.reset_mock()
                mockos.path.exists.return_value = True
                files = config.rattail_default_files('rattail')
                generic_default_files.assert_not_called()
                self.assertEqual(files, [quiet])


class TestMakeConfig(FileConfigTestCase):

    def test_files(self):
        generic = self.write_file('generic.conf', '')
        myfile = self.write_file('my.conf', '')

        # generic files by default
        myconfig = config.make_config(default_files=[generic])
        self.assertEqual(myconfig.files_read, [generic])

        # can specify single primary file
        myconfig = config.make_config(myfile, default_files=[generic])
        self.assertEqual(myconfig.files_read, [myfile])

        # can specify primary files as list
        myconfig = config.make_config([myfile], default_files=[generic])
        self.assertEqual(myconfig.files_read, [myfile])

        # can specify primary files via env
        myconfig = config.make_config(env={'RATTAIL_CONFIG_FILES': myfile},
                                      default_files=[generic])
        self.assertEqual(myconfig.files_read, [myfile])



# -*- coding: utf-8 -*-
"""add bouncer

Revision ID: 35b1da3be8f5
Revises: 3623367c499f
Create Date: 2015-07-21 18:12:24.027137

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '35b1da3be8f5'
down_revision = u'3623367c499f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # email_bounce
    op.create_table('email_bounce',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('config_key', sa.String(length=20), nullable=False),
                    sa.Column('bounced', sa.DateTime(), nullable=False),
                    sa.Column('bounce_recipient_address', sa.String(length=255), nullable=False),
                    sa.Column('intended_recipient_address', sa.String(length=255), nullable=True),
                    sa.Column('intended_recipient_key', sa.String(length=20), nullable=True),
                    sa.Column('processed', sa.DateTime(), nullable=True),
                    sa.Column('processed_by_uuid', sa.String(length=32), nullable=True),
                    sa.ForeignKeyConstraint(['processed_by_uuid'], [u'user.uuid'], name=u'email_bounce_fk_processed_by'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # email_bounce
    op.drop_table('email_bounce')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Installer utilities
"""

import os
import stat
import subprocess
import sys

from rattail.app import GenericHandler
from rattail.files import resource_path
from rattail.mako import ResourceTemplateLookup
from rattail.commands.util import require_prompt_toolkit, require_rich, rprint, basic_prompt


class InstallHandler(GenericHandler):
    """
    Base class and default implementation for ``poser install``
    commands.  Note that there is no ``rattail install`` command.
    """
    # nb. these must be explicitly set b/c config is not available
    # when running normally, e.g. `poser -n install`
    #app_title = "Poser"
    #app_package = 'poser'
    #app_eggname = 'Poser'
    #app_pypiname = 'Poser'

    def __init__(self, config,
                 app_title=None, app_package=None, app_eggname=None, app_pypiname=None,
                 main_image_url=None, header_image_url=None, favicon_url=None,
                 **kwargs):
        super().__init__(config, **kwargs)

        if app_title:
            self.app_title = app_title
        if app_package:
            self.app_package = app_package
        if app_eggname:
            self.app_eggname = app_eggname
        if app_pypiname:
            self.app_pypiname = app_pypiname

        self.main_image_url = main_image_url or '/tailbone/img/home_logo.png'
        self.header_image_url = header_image_url or '/tailbone/img/rattail.ico'
        self.favicon_url = favicon_url or '/tailbone/img/rattail.ico'

    def run(self):

        # nb. technically these would be required and auto-installed
        # as needed (later), but seems better to do this explicitly
        # up-front before any other command output
        require_prompt_toolkit()
        require_rich()

        self.templates = ResourceTemplateLookup(directories=[
            resource_path('{}:templates/installer'.format(self.app_package)),
            resource_path('rattail:templates/installer'),
        ])

        self.schema_installed = False
        self.do_install_steps()

        rprint("\n\t[bold green]initial setup is complete![/bold green]")

        self.show_goodbye()
        rprint()

    def do_install_steps(self):
        self.show_welcome()
        self.sanity_check()

        # prompt user for db info
        dbinfo = self.get_dbinfo()

        # get context for generated app files
        context = self.make_template_context(dbinfo)

        # make the appdir
        self.make_appdir(context)

        # install db schema if user likes
        self.schema_installed = self.install_db_schema(dbinfo)

    def show_welcome(self, **kwargs):

        rprint("\n\t[blue]Welcome to {}![/blue]".format(self.app_title))
        rprint("\n\tThis tool will install and configure a new app.")
        rprint("\n\t[italic]NB. You should already have created a new database in PostgreSQL or MySQL.[/italic]")

        # continue?
        if not basic_prompt("continue?", True, is_bool=True):
            rprint()
            sys.exit(0)

    def sanity_check(self, **kwargs):

        # appdir must not yet exist
        appdir = os.path.join(sys.prefix, 'app')
        if os.path.exists(appdir):
            rprint("\n\t[bold red]appdir already exists:[/bold red]  {}\n".format(appdir))
            sys.exit(1)

    def get_dbinfo(self, **kwargs):
        dbinfo = {}

        # get db info
        dbinfo['dbtype'] = basic_prompt('db type', 'postgresql')
        dbinfo['dbhost'] = basic_prompt('db host', 'localhost')
        default_port = '3306' if dbinfo['dbtype'] == 'mysql' else '5432'
        dbinfo['dbport'] = basic_prompt('db port', default_port)
        dbinfo['dbname'] = basic_prompt('db name', self.app_package)
        dbinfo['dbuser'] = basic_prompt('db user', 'rattail')

        # get db password
        dbinfo['dbpass'] = None
        while not dbinfo['dbpass']:
            dbinfo['dbpass'] = basic_prompt('db pass', is_password=True)

        # test db connection
        rprint("\n\ttesting db connection... ", end='')
        dbinfo['dburl'] = self.make_db_url(dbinfo['dbtype'],
                                           dbinfo['dbhost'],
                                           dbinfo['dbport'],
                                           dbinfo['dbname'],
                                           dbinfo['dbuser'],
                                           dbinfo['dbpass'])
        error = self.test_db_connection(dbinfo['dburl'])
        if error:
            rprint("[bold red]cannot connect![/bold red] ..error was:")
            rprint("\n{}".format(error))
            rprint("\n\t[bold yellow]aborting mission[/bold yellow]\n")
            sys.exit(1)
        rprint("[bold green]good[/bold green]")

        return dbinfo

    def make_db_url(self, dbtype, dbhost, dbport, dbname, dbuser, dbpass):
        try:
            # newer style
            from sqlalchemy.engine import URL
            factory = URL.create
        except ImportError:
            # older style
            from sqlalchemy.engine.url import URL
            factory = URL

        if dbtype == 'mysql':
            drivername = 'mysql+mysqlconnector'
        else:
            drivername = 'postgresql+psycopg2'

        return factory(drivername=drivername,
                       username=dbuser,
                       password=dbpass,
                       host=dbhost,
                       port=dbport,
                       database=dbname)

    def test_db_connection(self, url):
        from sqlalchemy import create_engine, inspect

        engine = create_engine(url)

        # check for random table; does not matter if it exists, we
        # just need to test interaction and this is a neutral way
        try:
            inspect(engine).has_table('whatever')
        except Exception as error:
            return str(error)

    def make_template_context(self, dbinfo, **kwargs):
        envname = os.path.basename(sys.prefix)
        appdir = os.path.join(sys.prefix, 'app')
        return {
            'envdir': sys.prefix,
            'envname': envname,
            'app_package': self.app_package,
            'app_title': self.app_title,
            'pypi_name': self.app_pypiname,
            'appdir': appdir,
            'db_url': dbinfo['dburl'],
            'pyramid_egg': self.app_eggname,
            'beaker_key': envname,
        }

    def make_appdir(self, context, **kwargs):
        rootpkg = self.app_package

        # appdir
        appdir = os.path.join(sys.prefix, 'app')
        self.app.make_appdir(appdir)

        # rattail.conf
        template = self.templates.get_template('rattail.conf.mako')
        self.app.make_config_file(
            'rattail', os.path.join(appdir, 'rattail.conf'),
            template=template, **context)

        # quiet.conf, silent.conf
        self.app.make_config_file('quiet', appdir)
        self.app.make_config_file('silent', appdir)

        # web.conf
        self.app.make_config_file(
            'web-complete', os.path.join(appdir, 'web.conf'),
            **context)

        # upgrade.sh
        template = self.templates.get_template('upgrade.sh.mako')
        path = os.path.join(appdir, 'upgrade.sh')
        self.app.render_mako_template(None, context, template=template,
                                      output_path=path)
        os.chmod(path, stat.S_IRWXU
                 | stat.S_IRGRP
                 | stat.S_IXGRP
                 | stat.S_IROTH
                 | stat.S_IXOTH)

        rprint("\n\tappdir created at:  [bold green]{}[/bold green]".format(appdir))

    def install_db_schema(self, dbinfo, **kwargs):
        from alembic.util.messaging import obfuscate_url_pw

        if not basic_prompt("install db schema?", True, is_bool=True):
            return False

        rprint()

        # install db schema
        cmd = [os.path.join(sys.prefix, 'bin', 'alembic'),
               '-c', os.path.join(sys.prefix, 'app', 'rattail.conf'),
               'upgrade', 'heads']
        subprocess.check_call(cmd)

        # put initial settings
        self.put_settings()

        rprint("\n\tdb schema installed to:  [bold green]{}[/bold green]".format(
            obfuscate_url_pw(dbinfo['dburl'])))

        # make admin, if user likes
        self.make_admin_user()

        return True

    def put_settings(self, **kwargs):

        # set falafel theme
        self.put_setting('tailbone.theme', 'falafel')

        # hide theme picker
        self.put_setting('tailbone.themes.expose_picker', 'false')

        # set main image
        self.put_setting('tailbone.main_image_url', self.main_image_url)

        # set header image
        self.put_setting('tailbone.header_image_url', self.header_image_url)

        # set favicon image
        self.put_setting('tailbone.favicon_url', self.favicon_url)

        # set default grid page size
        self.put_setting('tailbone.grid.default_pagesize', '20')

    def put_setting(self, name, value):
        cmd = [os.path.join(sys.prefix, 'bin', 'rattail'),
               '-c', os.path.join(sys.prefix, 'app', 'silent.conf'),
               'setting-put', name, value]
        subprocess.check_call(cmd)

    def make_admin_user(self, **kwargs):

        if not basic_prompt("create admin user?", True, is_bool=True):
            return False

        # get admin credentials
        username = basic_prompt('admin username', 'admin')
        password = None
        while not password:
            password = basic_prompt('admin password', is_password=True)
            if password:
                confirm = basic_prompt('confirm password', is_password=True)
                if not confirm or confirm != password:
                    rprint("[bold yellow]passwords did not match[/bold yellow]")
                    password = None
        fullname = basic_prompt('full name')

        rprint()

        # make admin user
        cmd = [os.path.join(sys.prefix, 'bin', 'rattail'),
               '-c', os.path.join(sys.prefix, 'app', 'quiet.conf'),
               'make-user', '--admin', username, '--password', password]
        if fullname:
            cmd.extend(['--full-name', fullname])
        subprocess.check_call(cmd)

        rprint("\n\tadmin user created:  [bold green]{}[/bold green]".format(
            username))
        return True

    def show_goodbye(self):
        if self.schema_installed:
            rprint("\n\tyou can run the web app with:")
            rprint("\n\t[blue]cd {}[/blue]".format(sys.prefix))
            rprint("\t[blue]bin/pserve file+ini:app/web.conf[/blue]")



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Configuration
"""

import logging
from collections import OrderedDict

from rattail.exceptions import SQLAlchemyNotInstalled
from rattail.config import parse_list


log = logging.getLogger(__name__)


def get_engines(config, section='rattail.db'):
    """
    Fetch all database engines defined within a given config section.

    This is a compatibility wrapper around
    :func:`wuttjamaican:wuttjamaican.db.conf.get_engines()`.  New code
    should use that instead; this will eventually be removed.
    """
    # nb. can only defer to wutta logic if we are using newer config,
    # because it assumes config does not need (section, option) and
    # can work with (key) instead
    if getattr(config, '_style', None) not in ('wuttaconfig', 'configuration'):

        # okay then we have legacy config...
        app = config.get_app()
        keys = config.get(section, 'keys', usedb=False)
        if keys:
            keys = parse_list(keys)
        else:
            keys = ['default']

        engines = OrderedDict()
        cfg = config.get_dict(section)
        for key in keys:
            key = key.strip()
            try:
                engines[key] = app.make_engine_from_config(cfg, prefix=f'{key}.')
            except KeyError:
                if key == 'default':
                    try:
                        engines[key] = app.make_engine_from_config(cfg, prefix='sqlalchemy.')
                    except KeyError:
                        pass

        return engines

    # but if config style is newer, use upstream logic.
    # nb. must not import this at module top in case sqlalchemy not installed
    from wuttjamaican.db import get_engines
    return get_engines(config, section)


# TODO: Deprecate/remove this.
def get_default_engine(config, section='rattail.db'):
    """
    Fetch the default database engine defined in the given config object for a
    given section.

    :param config: A ``ConfigParser`` instance containing app configuration.

    :type section: string
    :param section: Optional section name within which the configuration
       options are defined.  If not specified, ``'rattail.db'`` is assumed.

    :returns: A SQLAlchemy engine instance, or ``None``.

    .. note::
       This function calls :func:`get_engines()` for the heavy lifting; see
       that function for more details on how the engine configuration is read.
    """
    return get_engines(config, section=section).get('default')


# TODO: Deprecate/remove this.
def configure_session(config, session):
    """
    Configure a session factory or instance.  Currently all this does is
    install the hook to record changes, if config so dictates.
    """
    if config.getbool('rattail.db', 'changes.record', usedb=False):
        from rattail.db.changes import record_changes
        record_changes(session, config=config)


def configure_versioning(config, force=False, manager=None, plugins=None, **kwargs):
    """
    Configure Continuum versioning.
    """
    if not config.versioning_enabled() and not force:
        return

    try:
        from sqlalchemy.orm import configure_mappers
        import sqlalchemy_continuum as continuum
        from sqlalchemy_continuum.plugins import TransactionMetaPlugin
        from rattail.db.continuum import versioning_manager, RattailPlugin
    except ImportError as error:
        raise SQLAlchemyNotInstalled(error)
    else:
        kwargs['manager'] = manager or versioning_manager
        if plugins:
            kwargs['plugins'] = plugins
        else:
            kwargs['plugins'] = [TransactionMetaPlugin(), RattailPlugin()]
        log.debug("enabling Continuum versioning")
        continuum.make_versioned(**kwargs)

        # TODO: is this the best way/place to confirm versioning?
        try:
            model = config.get_model()
            configure_mappers()
            transaction_class = continuum.transaction_class(model.User)
            config.versioning_has_been_enabled = True
        except continuum.ClassNotVersioned:
            raise RuntimeError("Versioning is enabled and configured, but is not functional!  "
                               "This probably means the code import sequence is faulty somehow.  "
                               "Please investigate ASAP.")



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Telemetry Commands
"""

import logging
import pprint

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


log = logging.getLogger(__name__)


@rattail_typer.command()
def telemetry(
        ctx: typer.Context,
        profile: Annotated[
            str,
            typer.Option('--profile', '-p',
                         help="Profile (type) of telemetry data to collect.  "
                         "This also determines where/how data is submitted.  "
                         "If not specified, default logic is assumed.")] = None,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through all the motions but do not submit "
                         "the data to server.")] = False,
):
    """
    Submit telemetry data to a server
    """
    config = ctx.parent.rattail_config
    do_telemetry(config, profile, dry_run=dry_run, verbose=ctx.parent.params['verbose'])


class Telemetry(Subcommand):
    """
    Submit telemetry data to a server
    """
    name = 'telemetry'
    description = __doc__.strip()

    def add_parser_args(self, parser):

        parser.add_argument('--profile', '-p', metavar='KEY',
                            help="Profile (type) of telemetry data to collect.  "
                            "This also determines where/how data is submitted.  "
                            "If not specified, default logic is assumed.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through all the motions but do not submit "
                            "the data to server.")

    def run(self, args):
        do_telemetry(self.config, args.profile, dry_run=args.dry_run, verbose=self.verbose)


def do_telemetry(config, profile, dry_run=False, verbose=False):
    app = config.get_app()
    telemetry = app.get_telemetry_handler()

    data = telemetry.collect_all_data(profile=profile)
    log.info("data collected okay: %s", ', '.join(sorted(data)))
    log.debug("%s", data)

    if verbose:
        print("COLLECTED DATA:")
        pprint.pprint(data)

    if dry_run:
        log.info("dry run, so will not submit data to server")
    else:
        telemetry.submit_all_data(data, profile=profile)
        log.info("data submitted okay")



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Backup commands
"""

from pathlib import Path
from typing import List

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand
from rattail.backup import BackupHandler


@rattail_typer.command()
def backup(
        ctx: typer.Context,

        # db
        list_dbs: Annotated[
            bool,
            typer.Option('--list-dbs',
                         help="List all databases which may be backed up, then exit.")] = False,
        dbdump: Annotated[
            bool,
            typer.Option('--dbdump',
                         help="Dump some or all databases during the backup run.  Exactly "
                            "which databases are dumped, will depend on other parameters.")] = False,
        no_dbdump: Annotated[
            bool,
            typer.Option('--no-dbdump',
                         help="Do NOT dump any databases during the backup run.")] = False,
        included_dbs: Annotated[
            List[str],
            typer.Option('--dbdump-include',
                         help="Name of database to include in the dump.  May be specified "
                         "more than once.  Note that if this parameter is used, then ONLY "
                         "those databases specified will be backed up; the --dbdump-exclude "
                         "parameter(s) will be ignored.")] = None,
        excluded_dbs: Annotated[
            List[str],
            typer.Option('--dbdump-exclude',
                         help="Name of database to exclude from the dump.  May be specified "
                         "more than once.  Note that if the --dbdump-include parameter is "
                         "specified, then this parameter will be ignored.")] = None,
        dump_tables: Annotated[
            bool,
            typer.Option('--dump-tables',
                         help="Dump each table for specified databases, in addition to "
                         "the database as a whole.")] = False,
        no_dump_tables: Annotated[
            bool,
            typer.Option('--no-dump-tables',
                         help="Do *not* dump each table for each  database.")] = False,
        dbdump_output: Annotated[
            Path,
            typer.Option(help="Location of output folder for db dumps.")] = '/root/data',

        # rsync
        rsync: Annotated[
            bool,
            typer.Option('--rsync',
                         help="Push all files to remote location via rsync.")] = False,
        no_rsync: Annotated[
            bool,
            typer.Option('--no-rsync',
                         help="Do NOT push files to remote location via rsync.")] = False,
        rsync_included_prefixes: Annotated[
            List[str],
            typer.Option('--rsync-include',
                         help="File prefix which should be included in the rsync run.  "
                         "Actually, a separate rsync call is made for each prefix.  May "
                         "be specified more than once.  Note that if this parameter is "
                         "used, then ONLY those prefixes specified will be rsync'ed.  "
                         "Also, the --rsync-exclude parameter(s) will still apply.")] = None,
        rsync_excluded_prefixes: Annotated[
            List[str],
            typer.Option('--rsync-exclude',
                         help="File prefix which should be excluded from the rsync run.  "
                         "May be specified more than once.  Note that if the "
                         "--rsync-include parameter is specified, then this parameter(s) "
                         "will still apply also.")] = None,
        rsync_remote_host: Annotated[
            str,
            typer.Option(help="Assuming the rsync destination is another server, this should "
                         "be the hostname or IP address of that server.  Or you could set it "
                         "to an empty string, which signifies that the rsync remote "
                         "destination is another folder on localhost.")] = None,
        rsync_remote_prefix: Annotated[
            str,
            typer.Option(help="File prefix for the rsync destination.  If "
                         "--rsync-remote-host is an empty string, this prefix must exist "
                         "on localhost.  Otherwise the prefix must exist on the rsync "
                         "remote host.")] = None,

        # borg
        borg: Annotated[
            bool,
            typer.Option('--borg',
                         help="Create backup archive(s) via Borg.")] = False,
        no_borg: Annotated[
            bool,
            typer.Option('--no-borg',
                         help="Do NOT create backup archive(s) via Borg.")] = False,
        borg_remotes: Annotated[
            str,
            typer.Option(help="List (comma-separated string) of named \"remotes\" for "
                         "Borg, where archives should be created.")] = None,
        borg_included_prefixes: Annotated[
            List[str],
            typer.Option('--borg-include',
                         help="File prefix which should be included in the borg run.  "
                         "May be specified more than once.  If this is *not* specified "
                         "then --rsync-include will be used, if it is specified.  Note "
                         "that if this parameter is used, then ONLY those prefixes "
                         "specified will be included.  Also, the --borg-exclude "
                         "parameter(s) will still apply.")] = None,
        borg_excluded_prefixes: Annotated[
            List[str],
            typer.Option('--borg-exclude',
                         help="File prefix which should be excluded from the borg run.  "
                         "May be specified more than once.  If this is *not* specified "
                         "then --rsync-exclude will be used, if it is specified.  Note "
                         "that if the --borg-include parameter is specified, then this "
                         "parameter(s) will still apply also.")] = None,
        borg_tag: Annotated[
            str,
            typer.Option(help="Extra \"tag\" to embed within the name of the "
                         "Borg archive which is created.  Note that an archive "
                         "which is \"tagged\" in this way will never be pruned "
                         "by this `rattail backup` command!")] = None,

        # misc.
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the motions as much as possible, to get an idea "
                         "of what the full backup would do, but don't actually do anything.")] = False,
):
    """
    Backup the database(s) and/or files for this machine
    """
    config = ctx.parent.rattail_config

    params = dict(ctx.params)
    params['verbose'] = ctx.parent.params['verbose']

    handler = BackupHandler(config)
    handler.execute(params)


class Backup(Subcommand):
    """
    Backup the database(s) and/or files for this machine
    """
    name = 'backup'
    description = __doc__.strip()

    def add_args(self):
        """ """
        super().add_args()
        parser = self.parser

        parser.add_argument('--list-dbs', action='store_true',
                            help="List all databases which may be backed up, then exit.")

        parser.add_argument('--dbdump', action='store_true',
                            help="Dump some or all databases during the backup run.  Exactly "
                            "which databases are dumped, will depend on other parameters.")
        parser.add_argument('--no-dbdump', action='store_true',
                            help="Do NOT dump any databases during the backup run.")

        parser.add_argument('--dbdump-include', action='append', dest='included_dbs', metavar='NAME',
                            help="Name of database to include in the dump.  May be specified "
                            "more than once.  Note that if this parameter is used, then ONLY "
                            "those databases specified will be backed up. Also, the --dbdump-exclude "
                            "parameter(s) will be ignored.")
        parser.add_argument('--dbdump-exclude', action='append', dest='excluded_dbs', metavar='NAME',
                            help="Name of database to exclude from the dump.  May be specified "
                            "more than once.  Note that if the --dbdump-include parameter is specified, "
                            "then this parameter will be ignored.")

        parser.add_argument('--dbdump-output', default='/root/data', metavar='PATH',
                            help="Location of output folder for db dumps.  Default is /root/data")

        parser.add_argument('--dump-tables', action='store_true',
                            help="Dump each table for specified databases, in addition to the database "
                            "as a whole.")
        parser.add_argument('--no-dump-tables', action='store_true',
                            help="Do *not* dump each table for each  database.")

        parser.add_argument('--rsync', action='store_true',
                            help="Push all files to remote location via rsync.")
        parser.add_argument('--no-rsync', action='store_true',
                            help="Do NOT push files to remote location via rsync.")

        parser.add_argument('--rsync-include', action='append', dest='rsync_included_prefixes', metavar='PREFIX',
                            help="File prefix which should be included in the rsync run.  Actually, a "
                            "separate rsync call is made for each prefix.  May be specified more than "
                            "once.  Note that if this parameter is used, then ONLY those prefixes specified "
                            "will be rsync'ed.  Also, the --rsync-exclude parameter(s) will still apply.")
        parser.add_argument('--rsync-exclude', action='append', dest='rsync_excluded_prefixes', metavar='PREFIX',
                            help="File prefix which should be excluded from the rsync run.  May be "
                            "specified more than once.  Note that if the --rsync-include parameter is "
                            "specified, then this parameter(s) will still apply also.")

        parser.add_argument('--rsync-remote-host', metavar='ADDRESS',
                            help="Assuming the rsync destination is another server, this should be the "
                            "hostname or IP address of that server.  Or you could set it to an empty string, "
                            "which signifies that the rsync remote destination is another folder on localhost.")
        parser.add_argument('--rsync-remote-prefix', metavar='PREFIX',
                            help="File prefix for the rsync destination.  If --rsync-remote-host is an empty "
                            "string, this prefix must exist on localhost.  Otherwise the prefix must exist on "
                            "the rsync remote host.")

        parser.add_argument('--borg', action='store_true',
                            help="Create backup archive(s) via Borg.")
        parser.add_argument('--no-borg', action='store_true',
                            help="Do NOT create backup archive(s) via Borg.")

        parser.add_argument('--borg-remotes', metavar='REMOTES',
                            help="List of named \"remotes\" for Borg, where archives should be created.")

        parser.add_argument('--borg-include', action='append', dest='borg_included_prefixes', metavar='PREFIX',
                            help="File prefix which should be included in the borg run.  May be specified more "
                            "than once.  If this is *not* specified then --rsync-include will be used, if it is "
                            "specified.  Note that if this parameter is used, then ONLY those prefixes specified "
                            "will be included.  Also, the --borg-exclude parameter(s) will still apply.")
        parser.add_argument('--borg-exclude', action='append', dest='borg_excluded_prefixes', metavar='PREFIX',
                            help="File prefix which should be excluded from the borg run.  May be specified more "
                            "than once.  If this is *not* specified then --rsync-exclude will be used, if it is "
                            "specified.  Note that if the --borg-include parameter is specified, then this "
                            "parameter(s) will still apply also.")

        parser.add_argument('--borg-tag', metavar='TAG',
                            help="Extra \"tag\" to embed within the name of the "
                            "Borg archive which is created.  Note that an archive "
                            "which is \"tagged\" in this way will never be pruned "
                            "by this `rattail backup` command!")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the motions as much as possible, to get an idea "
                            "of what the full backup would do, but don't actually do anything.")

    def run(self, args):
        params = dict(args._get_kwargs())
        params['verbose'] = self.verbose

        handler = BackupHandler(self.config)
        handler.execute(params)



# -*- coding: utf-8; -*-
"""grow batch_pricing_row.margins

Revision ID: 04f17c46b42f
Revises: 7eb0903932ff
Create Date: 2018-11-26 18:04:14.446916

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '04f17c46b42f'
down_revision = '7eb0903932ff'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pricing_row
    op.alter_column('batch_pricing_row', 'true_margin', type_=sa.Numeric(precision=8, scale=3))
    op.alter_column('batch_pricing_row', 'price_margin', type_=sa.Numeric(precision=8, scale=3))
    op.alter_column('batch_pricing_row', 'margin_diff', type_=sa.Numeric(precision=8, scale=3))


def downgrade():

    # batch_pricing_row
    op.alter_column('batch_pricing_row', 'true_margin', type_=sa.Numeric(precision=6, scale=3))
    op.alter_column('batch_pricing_row', 'price_margin', type_=sa.Numeric(precision=6, scale=3))
    op.alter_column('batch_pricing_row', 'margin_diff', type_=sa.Numeric(precision=6, scale=3))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Command utilities
"""

import subprocess
import sys

from wuttjamaican.util import parse_bool


def require_prompt_toolkit():
    try:
        import prompt_toolkit
    except ImportError:
        value = input("\nprompt_toolkit is not installed.  shall i install it? [Yn] ")
        value = value.strip()
        if value and not parse_bool(value):
            sys.stderr.write("prompt_toolkit is required; aborting\n")
            sys.exit(1)

        subprocess.check_call([sys.executable, '-m', 'pip',
                               'install', 'prompt_toolkit'])


def require_rich():
    try:
        import rich
    except ImportError:
        value = input("\nrich is not installed.  shall i install it? [Yn] ")
        value = value.strip()
        if value and not parse_bool(value):
            sys.stderr.write("rich is required; aborting\n")
            sys.exit(1)

        subprocess.check_call([sys.executable, '-m', 'pip',
                               'install', 'rich'])


def rprint(*args, **kwargs):
    require_rich()

    # TODO: this could look different once python2 is out of the
    # picture; but must avoid `print` keyword for python2
    import rich
    rprint = getattr(rich, 'print')
    return rprint(*args, **kwargs)


def basic_prompt(info, default=None, is_password=False, is_bool=False, required=False):
    require_prompt_toolkit()

    from prompt_toolkit import prompt
    from prompt_toolkit.styles import Style

    # message formatting styles
    style = Style.from_dict({
        '': '',
        'bold': 'bold',
    })

    # build prompt message
    message = [
        ('', '\n'),
        ('class:bold', info),
    ]
    if default is not None:
        if is_bool:
            message.append(('', ' [{}]: '.format('Y' if default else 'N')))
        else:
            message.append(('', ' [{}]: '.format(default)))
    else:
        message.append(('', ': '))

    # prompt user for input
    try:
        text = prompt(message, style=style, is_password=is_password)
    except (KeyboardInterrupt, EOFError):
        rprint("\n\t[bold yellow]operation canceled by user[/bold yellow]\n",
                    file=sys.stderr)
        sys.exit(2)

    if is_bool:
        if text == '':
            return default
        elif text.upper() == 'Y':
            return True
        elif text.upper() == 'N':
            return False
        rprint("\n\t[bold yellow]ambiguous, please try again[/bold yellow]\n")
        return basic_prompt(info, default, is_bool=True)

    if required and not text and not default:
        return basic_prompt(info, default, is_password=is_password,
                            required=True)

    return text or default



# -*- coding: utf-8; -*-
"""add User.prevent_password_change

Revision ID: 524373deb98e
Revises: e206b9457091
Create Date: 2023-05-02 18:49:58.063276

"""

# revision identifiers, used by Alembic.
revision = '524373deb98e'
down_revision = 'e206b9457091'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # user
    op.add_column('user', sa.Column('prevent_password_change', sa.Boolean(), nullable=True))
    op.add_column('user_version', sa.Column('prevent_password_change', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # user
    op.drop_column('user_version', 'prevent_password_change')
    op.drop_column('user', 'prevent_password_change')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
``rattail make-config`` command
"""

import os
from pathlib import Path

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


@rattail_typer.command()
def make_config(
        ctx: typer.Context,
        list_types: Annotated[
            bool,
            typer.Option('--list-types', '-l',
                         help="List the types of config files this tool can generate.")] = False,
        config_type: Annotated[
            str,
            typer.Option('--type', '-T',
                         help="Type of config file to create; defaults to 'rattail' "
                         "which will generate 'rattail.conf'")] = 'rattail',
        output: Annotated[
            Path,
            typer.Option('--output', '-O',
                         help="Path where the config file is to be generated.  This can "
                         "be the full path including filename, or just the folder, in which "
                         "case the filename is inferred from 'type'.  Default is to current "
                         "working folder.")] = '.',
):
    """
    Generate stub config file(s) where you want them
    """
    config = ctx.parent.rattail_config
    app = config.get_app()

    if list_types:
        list_config_types(config)
        return

    template_path = app.find_config_template(config_type)
    if not template_path:
        sys.stderr.write(f"config template not found for type: {config_type}\n")
        sys.exit(1)

    output_path = os.path.abspath(output)
    if os.path.isdir(output_path):
        output_path = os.path.join(output_path, f'{config_type}.conf'.format(config_type))
    if os.path.exists(output_path):
        sys.stderr.write(f"ERROR! output file already exists: {output_path}\n")
        sys.exit(2)

    config_path = app.make_config_file(config_type, output_path,
                                       template_path=template_path)
    sys.stdout.write(f"Config file generated at: {config_path}\n")


class MakeConfig(Subcommand):
    """
    Generate stub config file(s) where you want them
    """
    name = 'make-config'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('-T', '--type', metavar='NAME', default='rattail',
                            help="Type of config file to create; defaults to 'rattail' "
                            "which will generate 'rattail.conf'")
        parser.add_argument('-O', '--output', metavar='PATH', default='.',
                            help="Path where the config file is to be generated.  This can "
                            "be the full path including filename, or just the folder, in which "
                            "case the filename is inferred from 'type'.  Default is to current "
                            "working folder.")
        parser.add_argument('--list-types', '-l', action='store_true',
                            help="List the types of config files this tool can "
                            "generate.")

    def run(self, args):
        if args.list_types:
            list_config_types(self.config)
            return

        template_path = self.app.find_config_template(args.type)
        if not template_path:
            self.stderr.write("config template not found for type: {}\n".format(args.type))
            sys.exit(1)

        output_path = os.path.abspath(args.output)
        if os.path.isdir(output_path):
            output_path = os.path.join(output_path, '{}.conf'.format(args.type))
        if os.path.exists(output_path):
            self.stderr.write("ERROR! output file already exists: {}\n".format(output_path))
            sys.exit(2)

        config_path = self.app.make_config_file(args.type, output_path,
                                                template_path=template_path)
        self.stdout.write("Config file generated at: {}\n".format(config_path))


def list_config_types(config):
    app = config.get_app()
    templates = app.get_all_config_templates()
    sys.stdout.write("CONFIG TEMPLATES:\n")
    sys.stdout.write("=========================\n")
    for name, path in templates.items():
        sys.stdout.write("{:25s} {}\n".format(name, path))
    sys.stdout.write("=========================\n")



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
``rattail clonedb`` command
"""

import importlib
import logging
from typing import List

import typer
from typing_extensions import Annotated

from .base import rattail_typer, Subcommand


log = logging.getLogger(__name__)


@rattail_typer.command()
def clonedb(
        ctx: typer.Context,
        source_engine: Annotated[
            str,
            typer.Argument(help="SQLAlchemy engine URL for the source database.")] = ...,
        target_engine: Annotated[
            str,
            typer.Argument(help="SQLAlchemy engine URL for the target database.")] = ...,
        model: Annotated[
            str,
            typer.Option(help="Dotted path of Python module which contains the data model.")] = 'rattail.db.model',
        classes: Annotated[
            List[str],
            typer.Argument(help="Model classes which should be cloned.  Possible values here "
                           "depends on which module contains the data model.  If no classes "
                           "are specified, all available will be cloned.")] = None,
):
    """
    Clone (supported) data from a source DB to a target DB
    """
    from sqlalchemy import create_engine, orm

    config = ctx.parent.rattail_config
    app = config.get_app()
    progress = ctx.parent.rattail_progress
    model = importlib.import_module(model)
    assert classes

    source_engine = create_engine(source_engine)
    target_engine = create_engine(target_engine)
    model.Base.metadata.drop_all(bind=target_engine)
    model.Base.metadata.create_all(bind=target_engine)

    Session = orm.sessionmaker()
    src_session = Session(bind=source_engine)
    dst_session = Session(bind=target_engine)

    for clsname in classes:
        log.info("cloning data for model: %s", clsname)
        cls = getattr(model, clsname)
        src_query = src_session.query(cls)
        count = src_query.count()
        log.debug("found %d %s records to clone", count, clsname)
        if not count:
            continue

        mapper = orm.class_mapper(cls)
        key_query = src_session.query(*mapper.primary_key)

        def process(key, i):
            src_instance = src_query.get(key)
            dst_session.merge(src_instance)
            dst_session.flush()

        app.progress_loop(process, key_query, progress,
                          message=f"Cloning data for model: {clsname}")

    src_session.close()
    dst_session.commit()
    dst_session.close()


class CloneDatabase(Subcommand):
    """
    Clone (supported) data from a source DB to a target DB
    """
    name = 'clonedb'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('source_engine',
                            help="SQLAlchemy engine URL for the source database.")
        parser.add_argument('target_engine',
                            help="SQLAlchemy engine URL for the target database.")
        parser.add_argument('-m', '--model', default='rattail.db.model',
                            help="Dotted path of Python module which contains the data model.")
        parser.add_argument('-C', '--classes', nargs='*',
                            help="Model classes which should be cloned.  Possible values here "
                            "depends on which module contains the data model.  If no classes "
                            "are specified, all available will be cloned.")

    def run(self, args):
        from sqlalchemy import create_engine, orm

        model = importlib.import_module(args.model)
        classes = args.classes
        assert classes

        source_engine = create_engine(args.source_engine)
        target_engine = create_engine(args.target_engine)
        model.Base.metadata.drop_all(bind=target_engine)
        model.Base.metadata.create_all(bind=target_engine)

        Session = orm.sessionmaker()
        src_session = Session(bind=source_engine)
        dst_session = Session(bind=target_engine)

        for clsname in classes:
            log.info("cloning data for model: %s", clsname)
            cls = getattr(model, clsname)
            src_query = src_session.query(cls)
            count = src_query.count()
            log.debug("found %d %s records to clone", count, clsname)
            if not count:
                continue

            mapper = orm.class_mapper(cls)
            key_query = src_session.query(*mapper.primary_key)

            prog = None
            if self.progress:
                prog = self.progress("Cloning data for model: {0}".format(clsname), count)
            for i, key in enumerate(key_query, 1):

                src_instance = src_query.get(key)
                dst_session.merge(src_instance)
                dst_session.flush()

                if prog:
                    prog.update(i)
            if prog:
                prog.destroy()

        src_session.close()
        dst_session.commit()
        dst_session.close()
